#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tw.unspent: Tracking wallet unspent outputs class for the MMGen suite
"""

from ..util import msg, suf
from ..obj import (
	ImmutableAttr,
	ListItemAttr,
	MMGenListItem,
	TwComment,
	CoinTxID,
	NonNegativeInt)
from ..addr import CoinAddr
from ..amt import CoinAmtChk
from .shared import TwMMGenID, TwLabel, get_tw_label
from .view import TwView

class TwUnspentOutputs(TwView):

	has_age = False
	can_group = False
	show_mmid = True
	hdr_lbl = 'tracked addresses'
	desc    = 'address balances'
	item_desc = 'address'
	item_desc_pl = 'addresses'
	no_rpcdata_errmsg = """
		No spendable outputs found!  Import addresses with balances into your
		watch-only wallet using 'mmgen-addrimport' and then re-run this program.
	"""
	update_widths_on_age_toggle = False
	print_output_types = ('detail',)
	mod_subpath = 'tw.unspent'
	dump_fn_pfx = 'balances'
	prompt_fs_in = [
		'Sort options: [a]mount, a[d]dr, [M]mgen addr, [r]everse',
		'Display options: show [m]mgen addr, r[e]draw screen',
		'View/Print: pager [v]iew, [w]ide pager view, [p]rint to file{s}',
		'Actions: [q]uit menu, [D]elete addr, add [l]abel, [R]efresh balance:']
	key_mappings = {
		'a':'s_amt',
		'd':'s_addr',
		'r':'s_reverse',
		'M':'s_twmmid',
		'm':'d_mmid',
		'e':'d_redraw',
		'p':'a_print_detail',
		'v':'a_view',
		'w':'a_view_detail',
		'l':'i_comment_add'}
	extra_key_mappings = {
		'D':'i_addr_delete',
		'R':'i_balance_refresh'}
	disp_spc = 3
	vout_w = 0

	class display_type(TwView.display_type):

		class squeezed(TwView.display_type.squeezed):
			cols = ('num', 'addr', 'mmid', 'comment', 'amt', 'amt2')

		class detail(TwView.display_type.detail):
			cols = ('num', 'addr', 'mmid', 'amt', 'amt2', 'comment')

	class MMGenTwUnspentOutput(MMGenListItem):
		valid_attrs = {'txid', 'vout', 'amt', 'amt2', 'comment', 'twmmid', 'addr', 'confs', 'skip'}
		invalid_attrs = {'proto'}
		txid    = ListItemAttr(CoinTxID)
		vout    = ListItemAttr(NonNegativeInt)
		amt     = ImmutableAttr(CoinAmtChk, include_proto=True)
		amt2    = ListItemAttr(CoinAmtChk, include_proto=True) # the ETH balance for token account
		comment = ListItemAttr(TwComment, reassign_ok=True)
		twmmid  = ImmutableAttr(TwMMGenID, include_proto=True)
		addr    = ImmutableAttr(CoinAddr, include_proto=True)
		confs   = ImmutableAttr(int, typeconv=False)
		skip    = ListItemAttr(str, typeconv=False, reassign_ok=True)

		def __init__(self, proto, **kwargs):
			self.__dict__['proto'] = proto
			MMGenListItem.__init__(self, **kwargs)

	async def __init__(self, cfg, proto, *, minconf=1, addrs=[]):
		await super().__init__(cfg, proto)
		self.minconf  = NonNegativeInt(minconf)
		self.addrs    = addrs
		from ..cfg import gc
		self.min_cols = gc.min_screen_width

	@property
	def total(self):
		return sum(i.amt for i in self.data)

	def gen_data(self, rpc_data, lbl_id):
		for o in rpc_data:
			if not lbl_id in o:
				continue # coinbase outputs have no account field
			l = get_tw_label(self.proto, o[lbl_id])
			if l:
				if not 'amt' in o:
					o['amt'] = self.proto.coin_amt(o['amount'])
				o.update({
					'twmmid':  l.mmid,
					'comment': l.comment or '',
					'addr':    CoinAddr(self.proto, o['address']),
					'confs':   o['confirmations']
				})
				yield self.MMGenTwUnspentOutput(
					self.proto,
					**{k:v for k, v in o.items() if k in self.MMGenTwUnspentOutput.valid_attrs})

	async def get_rpc_data(self):
		wl = self.twctl.sorted_list
		minconf = int(self.minconf)
		block = self.twctl.rpc.get_block_from_minconf(minconf)
		if self.addrs:
			wl = [d for d in wl if d['addr'] in self.addrs]
		return [{
				'account': TwLabel(self.proto, d['mmid']+' '+d['comment']),
				'address': d['addr'],
				'amt': await self.twctl.get_balance(d['addr'], block=block),
				'confirmations': minconf,
				} for d in wl]

	def filter_data(self):

		data = self.data.copy()

		for d in data:
			d.skip = ''

		gkeys = {'addr': 'addr', 'twmmid': 'addr', 'txid': 'txid'}
		if self.group and self.sort_key in gkeys:
			for a, b in [(data[i], data[i+1]) for i in range(len(data)-1)]:
				for k in gkeys:
					if self.sort_key == k and getattr(a, k) == getattr(b, k):
						b.skip = gkeys[k]

		return data

	def get_column_widths(self, data, *, wide, interactive):

		show_mmid = self.show_mmid or wide

		return self.compute_column_widths(
			widths = { # fixed cols
				'num': max(2, len(str(len(data)))+1),
				'txid': 0,
				'vout': self.vout_w,
				'mmid': max(len(d.twmmid.disp) for d in data) if show_mmid else 0,
				'amt': self.amt_widths['amt'],
				'amt2': self.amt_widths.get('amt2', 0),
				'block': self.age_col_params['block'][0] if wide else 0,
				'date_time': self.age_col_params['date_time'][0] if wide else 0,
				'date': self.age_w,
				'spc': self.disp_spc + (2 * show_mmid) + self.has_amt2
			},
			maxws = { # expandable cols
				'addr': max(len(d.addr) for d in data),
				'comment': max(d.comment.screen_width for d in data) if show_mmid else 0,
			} | self.txid_max_w,
			minws = {
				'addr': 10,
				'comment': len('Comment') if show_mmid else 0,
			} | self.txid_min_w,
			maxws_nice = (
				self.nice_addr_w if show_mmid else {}
			) | self.txid_nice_w,
			wide = wide,
			interactive = interactive,
		)

	def squeezed_col_hdr(self, cw, fs, color):
		return fs.format(
			n = '',
			t = 'TxID',
			v = 'Vout',
			a = 'Address',
			m = 'MMGenID',
			c = 'Comment',
			A = 'Amt({})'.format(self.proto.dcoin),
			B = 'Amt({})'.format(self.proto.coin),
			d = self.age_hdr)

	def detail_col_hdr(self, cw, fs, color):
		return fs.format(
			n = '',
			t = 'TxID',
			v = 'Vout',
			a = 'Address',
			m = 'MMGenID',
			A = 'Amt({})'.format(self.proto.dcoin),
			B = 'Amt({})'.format(self.proto.coin),
			b = 'Block',
			D = 'Date/Time',
			c = 'Comment')

	def gen_squeezed_display(self, data, cw, fs, color, fmt_method):

		for n, d in enumerate(data):
			yield fs.format(
				n = str(n+1) + ')',
				t = (d.txid.fmtc('|' + '.'*(cw.txid-1), cw.txid, color=color) if d.skip  == 'txid'
					else d.txid.truncate(cw.txid, color=color)) if cw.txid else None,
				v = ' ' + d.vout.fmt(cw.vout-1, color=color) if cw.vout else None,
				a = d.addr.fmtc('|' + '.'*(cw.addr-1), cw.addr, color=color) if d.skip == 'addr'
					else d.addr.fmt(self.addr_view_pref, cw.addr, color=color),
				m = (d.twmmid.fmtc('.'*cw.mmid, cw.mmid, color=color) if d.skip == 'addr'
					else d.twmmid.fmt(cw.mmid, color=color)) if cw.mmid else None,
				c = d.comment.fmt2(cw.comment, color=color, nullrepl='-') if cw.comment else None,
				A = d.amt.fmt(cw.iwidth, color=color, prec=self.disp_prec),
				B = d.amt2.fmt(cw.iwidth2, color=color, prec=self.disp_prec) if cw.amt2 else None,
				d = self.age_disp(d, self.age_fmt),
			)

	def gen_detail_display(self, data, cw, fs, color, fmt_method):

		for n, d in enumerate(data):
			yield fs.format(
				n = str(n+1) + ')',
				t = d.txid.fmt(cw.txid, color=color) if cw.txid else None,
				v = ' ' + d.vout.fmt(cw.vout-1, color=color) if cw.vout else None,
				a = d.addr.fmt(self.addr_view_pref, cw.addr, color=color),
				m = d.twmmid.fmt(cw.mmid, color=color),
				A = d.amt.fmt(cw.iwidth, color=color, prec=self.disp_prec),
				B = d.amt2.fmt(cw.iwidth2, color=color, prec=self.disp_prec) if cw.amt2 else None,
				b = self.age_disp(d, 'block'),
				D = self.age_disp(d, 'date_time'),
				c = d.comment.fmt2(cw.comment, color=color, nullrepl='-'))

	def display_total(self):
		msg('\nTotal unspent: {} {} ({} {}{})'.format(
			self.total.hl(),
			self.proto.dcoin,
			len(self.data),
			self.item_desc,
			suf(self.data)))

	async def set_dates(self, us):
		if not self.dates_set:
			# 'blocktime' differs from 'time', is same as getblockheader['time']
			dates = [o.get('blocktime', 0)
				for o in await self.rpc.gathered_icall('gettransaction', [(o.txid, True, False) for o in us])]
			for idx, o in enumerate(us):
				o.date = dates[idx]
			self.dates_set = True

	class sort_action(TwView.sort_action):

		def s_twmmid(self, parent):
			parent.do_sort('twmmid')
			parent.show_mmid = True

	class display_action(TwView.display_action):

		def d_mmid(self, parent):
			parent.show_mmid = not parent.show_mmid

		def d_group(self, parent):
			if parent.can_group:
				parent.group = not parent.group



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
rpc.backends.aiohttp: aiohttp RPC backend for the MMGen Project
"""

import json

from ...base_obj import AsyncInit

from ..util import dmsg_rpc_backend, json_encoder

from .base import base

class aiohttp(base, metaclass=AsyncInit):
	"""
	Contrary to the requests library, aiohttp won’t read environment variables by
	default.  But you can do so by passing trust_env=True into aiohttp.ClientSession
	constructor to honor HTTP_PROXY, HTTPS_PROXY, WS_PROXY or WSS_PROXY environment
	variables (all are case insensitive).
	"""

	def __del__(self):
		self.connector.close()
		self.session.detach()
		del self.session

	async def __init__(self, caller):
		super().__init__(caller)
		import aiohttp
		self.connector = aiohttp.TCPConnector(limit_per_host=self.cfg.aiohttp_rpc_queue_len)
		self.session = aiohttp.ClientSession(
			headers = {'Content-Type': 'application/json'},
			connector = self.connector,
		)
		if caller.auth_type == 'basic':
			self.auth = aiohttp.BasicAuth(*caller.auth, encoding='UTF-8')
		else:
			self.auth = None

	async def run(self, payload, timeout, host_path):
		dmsg_rpc_backend(self.host_url, host_path, payload)
		async with self.session.post(
			url     = self.host_url + host_path,
			auth    = self.auth,
			data    = json.dumps(payload, cls=json_encoder),
			timeout = timeout or self.timeout,
		) as res:
			return (await res.text(), res.status)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.__init__: Monero wallet ops for the MMGen Suite
"""

import re, importlib
from collections import namedtuple

from ..proto.btc.common import b58a

from ..util import capfirst

tx_priorities = {
	1: 'low',
	2: 'normal',
	3: 'high',
	4: 'highest'
}

uargs = namedtuple('xmrwallet_uargs', [
	'infile',
	'wallets',
	'spec',
])

uarg_info = (
	lambda e, hp: {
		'daemon':          e('HOST:PORT', hp),
		'tx_relay_daemon': e('HOST:PORT[:PROXY_IP:PROXY_PORT]',     rf'({hp})(?::({hp}))?'),
		'newaddr_spec':    e('WALLET[:ACCOUNT][,"label text"]',     r'(\d+)(?::(\d+))?(?:,(.*))?'),
		'transfer_spec':   e('SOURCE:ACCOUNT:ADDRESS,AMOUNT',       rf'(\d+):(\d+):([{b58a}]+),([0-9.]+)'),
		'sweep_spec':      e('SOURCE:ACCOUNT[,DEST[:ACCOUNT]]',     r'(\d+):(\d+)(?:,(\d+)(?::(\d+))?)?'),
		'label_spec':      e('WALLET:ACCOUNT:ADDRESS,"label text"', r'(\d+):(\d+):(\d+),(.*)'),
	})(
		namedtuple('uarg_info_entry', ['annot','pat']),
		r'(?:[^:]+):(?:\d+)'
	)

# canonical op names mapped to their respective modules:
op_names = {
	'create':              'create',
	'create_offline':      'create',
	'sync':                'sync',
	'list':                'view',
	'view':                'view',
	'listview':            'view',
	'new':                 'new',
	'transfer':            'sweep',
	'sweep':               'sweep',
	'sweep_all':           'sweep',
	'relay':               'relay',
	'txview':              'txview',
	'txlist':              'txview',
	'label':               'label',
	'sign':                'sign',
	'submit':              'submit',
	'resubmit':            'submit',
	'abort':               'submit',
	'dump':                'dump',
	'restore':             'restore',
	'export_outputs':      'export',
	'export_outputs_sign': 'export',
	'import_outputs':      'import',
	'import_key_images':   'import',
	'wallet':              'wallet', # virtual class
}

kafile_arg_ops = (
	'create',
	'sync',
	'list',
	'view',
	'listview',
	'label',
	'new',
	'transfer',
	'sweep',
	'sweep_all',
	'dump',
	'restore')

opts = (
	'wallet_dir',
	'daemon',
	'tx_relay_daemon',
	'use_internal_keccak_module',
	'hash_preset',
	'restore_height',
	'no_start_wallet_daemon',
	'no_stop_wallet_daemon',
	'no_relay',
	'watch_only',
	'autosign',
	'skip_empty_accounts',
	'skip_empty_addresses')

pat_opts = ('daemon', 'tx_relay_daemon')

def op_cls(op_name):
	def upper(m):
		return m[1].upper()
	clsname = 'Op' + capfirst(re.sub(r'_(.)', upper, op_name))
	cls = getattr(importlib.import_module(f'.ops.{op_names[op_name]}', 'mmgen.xmrwallet'), clsname)
	cls.name = op_name
	return cls

def op(op, cfg, infile, wallets, *, spec=None):
	return op_cls(op.replace('-', '_'))(cfg, uargs(infile, wallets, spec))



#!/usr/bin/env python3

"""
test.modtest_d.indexed_dict: IndexedDict class unit test for the MMGen suite
"""

from mmgen.util import msg, msg_r, die

from ..include.common import vmsg

class unit_test:

	def run_test(self, name, ut):
		bad_msg = (
			'initializing values via constructor',
			'reassignment to existing key',
			'item deletion',
			'item moving',
			'clearing',
			'updating')
		def bad0(): IndexedDict(arg)
		def bad1(): d['a'] = 2
		def bad2(): del d['a']
		def bad3(): d.move_to_end('a')
		def bad4(): d.clear()
		def bad5(): d.update(d)

		def odie(n): die(4, f'\nillegal action {bad_msg[n]!r} failed to raise exception')
		def omsg(e): vmsg(' - ' + e.args[0])

		msg_r('Testing class IndexedDict...')

		from mmgen.obj import IndexedDict
		d = IndexedDict()

		d['a'] = 1
		d['b'] = 2

		vmsg('\nChecking error handling:')

		arg = [('a', 1), ('b', 2)]
		dict(arg)

		for n, func in enumerate([bad0, bad1, bad2, bad3, bad4, bad5]):
			try:
				func()
			except NotImplementedError as e:
				omsg(e)
			else:
				odie(n)

		try:
			d.key(2)
		except Exception as e:
			omsg(e)
		else:
			odie('list index out of range')

		d['c'] = 3

		d_chk = {'a':1, 'b':2, 'c':3}
		assert d == d_chk, d

		d_keys_chk = ['a', 'b', 'c']
		assert d.keys == d_keys_chk, d.keys

		A = d.key(0)
		assert A == 'a', A

		A = d.key(2)
		assert A == 'c', A

		msg('OK')
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.__init__: Monero wallet ops for the MMGen Suite
"""

import re, atexit

from ...color import blue
from ...util import msg, die, fmt
from ...protocol import init_proto
from ...tx.util import get_autosign_obj

from ... import xmrwallet

from .. import uarg_info

class OpBase:

	opts = ('wallet_dir',)
	trust_monerod = False
	do_umount = True
	name = None

	def __init__(self, cfg, uarg_tuple):

		def gen_classes():
			for cls in type(self).__mro__:
				if not cls.__name__.startswith('OpMixin'):
					yield cls
				if cls.__name__ == 'OpBase':
					break

		self.cfg = cfg
		classes = tuple(gen_classes())
		self.opts = tuple(set(opt for cls in classes for opt in xmrwallet.opts))

		if not hasattr(self, 'stem'):
			self.stem = self.name

		global fmt_amt, hl_amt, addr_width

		self.uargs = uarg_tuple

		def fmt_amt(amt):
			return self.proto.coin_amt(amt, from_unit='atomic').fmt(5, prec=12, color=True)
		def hl_amt(amt):
			return self.proto.coin_amt(amt, from_unit='atomic').hl()

		addr_width = 95 if self.cfg.full_address else 17

		self.proto = init_proto(cfg, 'xmr', network=self.cfg.network, need_amt=True)

		id_cur = None
		for cls in classes:
			if id(cls.check_uopts) != id_cur:
				cls.check_uopts(self)
				id_cur = id(cls.check_uopts)

		id_cur = None
		for cls in classes:
			if id(cls.pre_init_action) != id_cur:
				cls.pre_init_action(self)
				id_cur = id(cls.pre_init_action)

		if cfg.autosign:
			self.asi = get_autosign_obj(cfg)

	def check_uopts(self):

		def check_pat_opt(name):
			val = getattr(self.cfg, name)
			if not re.fullmatch(uarg_info[name].pat, val, re.ASCII):
				die(1, '{!r}: invalid value for --{}: it must have format {!r}'.format(
					val,
					name.replace('_', '-'),
					uarg_info[name].annot
				))

		for attr in self.cfg.__dict__:
			if attr in xmrwallet.opts and not attr in self.opts:
				die(1, 'Option --{} not supported for {!r} operation'.format(
					attr.replace('_', '-'),
					self.name,
				))

		for opt in xmrwallet.pat_opts:
			if getattr(self.cfg, opt, None):
				check_pat_opt(opt)

	def parse_tx_relay_opt(self):
		return re.fullmatch(
			uarg_info['tx_relay_daemon'].pat,
			self.cfg.tx_relay_daemon,
			re.ASCII)

	def display_tx_relay_info(self, *, indent=''):
		m = self.parse_tx_relay_opt()
		msg(fmt(f"""
			TX relay info:
			  Host:  {blue(m[1])}
			  Proxy: {blue(m[2] or 'None')}
			""", strip_char='\t', indent=indent))

	def mount_removable_device(self):
		if self.cfg.autosign:
			if not self.asi.device_inserted:
				die(1, 'Removable device not present!')
			if self.do_umount:
				atexit.register(lambda: self.asi.do_umount())
			self.asi.do_mount()
			self.post_mount_action()

	def pre_init_action(self):
		pass

	def post_main_success(self):
		pass

	def post_main_failure(self):
		pass

	async def stop_wallet_daemon(self):
		pass

	def post_mount_action(self):
		pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
main_wallet: Entry point for MMGen wallet-related scripts
"""

import sys, os
from .cfg import gc, Config
from .color import green, yellow
from .util import msg, gmsg_r, ymsg, bmsg, die, capfirst
from .wallet import Wallet, get_wallet_cls

usage = '[opts] [infile]'
nargs = 1
iaction = 'convert'
oaction = 'convert'
do_bw_note = True
do_sw_note = False
do_ss_note = False

invoked_as = {
	'mmgen-walletgen':    'gen',
	'mmgen-walletconv':   'conv',
	'mmgen-walletchk':    'chk',
	'mmgen-passchg':      'passchg',
	'mmgen-subwalletgen': 'subgen',
	'mmgen-seedsplit':    'seedsplit',
}[gc.prog_name]

dsw = f'the default or specified {gc.proj_name} wallet'

if invoked_as == 'gen':
	desc = f'Generate an {gc.proj_name} wallet from a random seed'
	usage = '[opts]'
	oaction = 'output'
	nargs = 0
elif invoked_as == 'conv':
	desc = 'Convert ' + dsw + ' from one format to another'
elif invoked_as == 'chk':
	desc = 'Check validity of ' + dsw
	iaction = 'input'
elif invoked_as == 'passchg':
	desc = 'Change the passphrase, hash preset or label of ' + dsw
	iaction = 'input'
	do_bw_note = False
elif invoked_as == 'subgen':
	desc = 'Generate a subwallet from ' + dsw
	usage = '[opts] [infile] <Subseed Index>'
	iaction = 'input'
	oaction = 'output'
	do_sw_note = True
elif invoked_as == 'seedsplit':
	desc = 'Generate a seed share from ' + dsw
	usage = '[opts] [infile] [<Split ID String>:]<index>:<share count>'
	iaction = 'input'
	oaction = 'output'
	do_ss_note = True

opts_data = {
	'filter_codes': {
		# Write  In-fmt  Out-fmt  Keep-pass  Force-update  Master-share  passwd-file-New-only
		'chk':       ['-',      'i'],
		'conv':      ['-', 'w', 'i', 'o', 'k',      'n'],
		'gen':       ['-', 'w',      'o'],
		'passchg':   ['-', 'w', 'i',      'k', 'f', 'n'],
		'seedsplit': ['-', 'w', 'i', 'o',      'm', 'n'],
		'subgen':    ['-', 'w', 'i', 'o', 'k',      'n'],
	}[invoked_as],
	'text': {
		'desc': desc,
		'usage': usage,
		'options': """
			-- -h, --help            Print this help message
			-- --, --longhelp        Print help message for long (global) options
			-w -d, --outdir=      d  Output files to directory 'd' instead of working dir
			-- -e, --echo-passphrase Echo passphrases and other user input to screen
			-f -f, --force-update    Force update of wallet even if nothing has changed
			-i -i, --in-fmt=      f  {iaction} from wallet format 'f' (see FMT CODES below)
			-o -o, --out-fmt=     f  {oaction} to wallet format 'f' (see FMT CODES below)
			-i -H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
			+                        'f' at offset 'o' (comma-separated)
			-o -J, --hidden-incog-output-params=f,o  Write hidden incognito data to file
			+                        'f' at offset 'o' (comma-separated). File 'f' will be
			+                        created if necessary and filled with random data.
			-i -O, --old-incog-fmt   Specify old-format incognito input
			-k -k, --keep-passphrase Reuse passphrase of input wallet for output wallet
			-k -K, --keep-hash-preset Reuse hash preset of input wallet for output wallet
			-- -l, --seed-len=    l  Specify wallet seed length of 'l' bits.  This option
			+                        is required only for brainwallet and incognito inputs
			+                        with non-standard (< {dsl}-bit) seed lengths.
			-w -L, --label=       l  Specify a label 'l' for output wallet
			-k -m, --keep-label      Reuse label of input wallet for output wallet
			-m -M, --master-share=i  Use a master share with index 'i' (min:{ms_min}, max:{ms_max})
			-- -p, --hash-preset= p  Use the scrypt hash parameters defined by preset 'p'
			+                        for password hashing (default: '{gc.dfl_hash_preset}')
			-- -z, --show-hash-presets Show information on available hash presets
			-- -P, --passwd-file= f  Get wallet passphrase from file 'f'
			-n -N, --passwd-file-new-only Use passwd file only for new, not existing, wallet
			-- -q, --quiet           Produce quieter output; suppress some warnings
			-- -r, --usr-randchars=n Get 'n' characters of additional randomness from user
			+                        (min={cfg.min_urandchars}, max={cfg.max_urandchars}, default={cfg.usr_randchars})
			-w -S, --stdout          Write wallet data to stdout instead of file
			-- -v, --verbose         Produce more verbose output
		""",
		'notes': """

{n_ss}{n_sw}{n_pw}{n_bw}

{f}
"""
	},
	'code': {
		'options': lambda cfg, help_notes, s: s.format(
			iaction = capfirst(iaction),
			oaction = capfirst(oaction),
			ms_min  = help_notes('MasterShareIdx').min_val,
			ms_max  = help_notes('MasterShareIdx').max_val,
			dsl     = help_notes('dfl_seed_len'),
			cfg     = cfg,
			gc      = gc,
		),
		'notes': lambda cfg, help_mod, help_notes, s: s.format(
			f       = help_notes('fmt_codes'),
			n_ss    = ('', help_mod('seedsplit')+'\n\n')[do_ss_note],
			n_sw    = ('', help_mod('subwallet')+'\n\n')[do_sw_note],
			n_pw    = help_notes('passwd'),
			n_bw    = ('', '\n\n'+help_notes('brainwallet'))[do_bw_note]
		)
	}
}

cfg = Config(opts_data=opts_data, need_proto=False)

cmd_args = cfg._args

if invoked_as == 'subgen':
	from .subseed import SubSeedIdx
	ss_idx = SubSeedIdx(cmd_args.pop())
elif invoked_as == 'seedsplit':
	from .obj import get_obj
	from .seedsplit import SeedSplitSpecifier, MasterShareIdx
	master_share = MasterShareIdx(cfg.master_share) if cfg.master_share else None
	if cmd_args:
		sss = get_obj(SeedSplitSpecifier, s=cmd_args.pop(), silent=True)
		if master_share:
			if not sss:
				sss = SeedSplitSpecifier('1:2')
			elif sss.idx == 1:
				m1 = 'Share index of 1 meaningless in master share context.'
				m2 = 'To generate a master share, omit the seed split specifier.'
				die(1, m1+'  '+m2)
		elif not sss:
			cfg._usage()
	elif master_share:
		sss = SeedSplitSpecifier('1:2')
	else:
		cfg._usage()

from .fileutil import check_infile, get_seed_file

if cmd_args:
	if invoked_as == 'gen' or len(cmd_args) > 1:
		cfg._usage()
	check_infile(cmd_args[0])

sf = get_seed_file(cfg, nargs=nargs, invoked_as=invoked_as)

if invoked_as != 'chk':
	from .ui import do_license_msg
	do_license_msg(cfg)

if invoked_as == 'gen':
	ss_in = None
else:
	ss_in = Wallet(
		cfg     = cfg,
		fn      = sf,
		passchg = invoked_as == 'passchg',
		passwd_file = False if cfg.passwd_file_new_only else None)
	m1 = green('Processing input wallet ')
	m2 = ss_in.seed.sid.hl()
	m3 = yellow(' (default wallet)') if sf and os.path.dirname(sf) == cfg.data_dir else ''
	msg(m1+m2+m3)

if invoked_as == 'chk':
	lbl = ss_in.ssdata.label.hl() if hasattr(ss_in.ssdata, 'label') else 'NONE'
	cfg._util.vmsg(f'Wallet label: {lbl}')
	# TODO: display creation date
	sys.exit(0)

if invoked_as != 'gen':
	gmsg_r('Processing output wallet' + ('\n', ' ')[invoked_as == 'seedsplit'])

if invoked_as == 'subgen':
	ss_out = Wallet(
		cfg      = cfg,
		seed_bin = ss_in.seed.subseed(ss_idx, print_msg=True).data)
elif invoked_as == 'seedsplit':
	shares = ss_in.seed.split(sss.count, id_str=sss.id, master_idx=master_share)
	seed_out = shares.get_share_by_idx(sss.idx, base_seed=True)
	msg(seed_out.get_desc(ui=True))
	ss_out = Wallet(
		cfg  = cfg,
		seed = seed_out)
else:
	ss_out = Wallet(
		cfg     = cfg,
		ss      = ss_in,
		passchg = invoked_as == 'passchg')

if invoked_as == 'gen':
	cfg._util.qmsg(f"This wallet's Seed ID: {ss_out.seed.sid.hl()}")

if invoked_as == 'passchg':
	def data_changed(attrs):
		for attr in attrs:
			if getattr(ss_out.ssdata, attr) != getattr(ss_in.ssdata, attr):
				return True
		return False
	if not (cfg.force_update or data_changed(('passwd', 'hash_preset', 'label'))):
		die(1, 'Password, hash preset and label are unchanged.  Taking no action')

if invoked_as == 'passchg':

	def secure_delete(fn):
		bmsg('Securely deleting old wallet')
		from .fileutil import shred_file
		shred_file(cfg, fn)

	def rename_old_wallet_maybe(silent):
		# though very unlikely, old and new wallets could have same Key ID and thus same filename.
		# If so, rename old wallet file before deleting.
		old_fn = ss_in.infile.name
		if os.path.basename(old_fn) == ss_out._filename():
			if not silent:
				ymsg(
					'Warning: diverting old wallet {old_fn!r} due to Key ID collision.  ' +
					'Please securely delete or move the diverted file!')
			os.rename(old_fn, old_fn+'.divert')
			return old_fn+'.divert'
		else:
			return old_fn

	if ss_in.infile.dirname == cfg.data_dir:
		from .ui import confirm_or_raise
		confirm_or_raise(
			cfg      = cfg,
			message  = yellow('Confirmation of default wallet update'),
			action   = 'update the default wallet',
			exit_msg = 'Password not changed')
		old_wallet = rename_old_wallet_maybe(silent=True)
		ss_out.write_to_file(desc='New wallet', outdir=cfg.data_dir)
		secure_delete(old_wallet)
	else:
		old_wallet = rename_old_wallet_maybe(silent=False)
		ss_out.write_to_file()
		from .ui import keypress_confirm
		if keypress_confirm(cfg, f'Securely delete old wallet {old_wallet!r}?'):
			secure_delete(old_wallet)
elif invoked_as == 'gen' and not cfg.outdir and not cfg.stdout:
	from .filename import find_file_in_dir
	if find_file_in_dir(get_wallet_cls('mmgen'), cfg.data_dir):
		ss_out.write_to_file()
	else:
		from .ui import keypress_confirm
		if keypress_confirm(
				cfg,
				'Make this wallet your default and move it to the data directory?',
				default_yes = True):
			ss_out.write_to_file(outdir=cfg.data_dir)
		else:
			ss_out.write_to_file()
else:
	ss_out.write_to_file()

if invoked_as == 'passchg':
	if ss_out.ssdata.passwd == ss_in.ssdata.passwd:
		msg('New and old passphrases are the same')
	else:
		msg('Wallet passphrase has changed')
	if ss_out.ssdata.hash_preset != ss_in.ssdata.hash_preset:
		msg(f'Hash preset has been changed to {ss_out.ssdata.hash_preset!r}')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.fileutil: File routines for the 'mmgen-tool' utility
"""

import sys, os

from .common import tool_cmd_base
from ..util import msg, msg_r, die, suf, make_full_path
from ..crypto import Crypto

class tool_cmd(tool_cmd_base):
	"file utilities"

	def find_incog_data(self,
			filename: str,
			incog_id: str,
			*,
			keep_searching: 'continue search after finding data (ID collisions can yield false positives)' = False):
		"Use an Incog ID to find hidden incognito wallet data"

		from hashlib import sha256

		ivsize, bsize, mod = (Crypto.aesctr_iv_len, 4096, 4096*8)
		n, carry = 0, b' '*ivsize
		flgs = os.O_RDONLY|os.O_BINARY if sys.platform == 'win32' else os.O_RDONLY
		f = os.open(filename, flgs)
		for ch in incog_id:
			if ch not in '0123456789ABCDEF':
				die(2, f'{incog_id!r}: invalid Incog ID')
		while True:
			d = os.read(f, bsize)
			if not d:
				break
			d = carry + d
			for i in range(bsize):
				if sha256(d[i:i+ivsize]).hexdigest()[:8].upper() == incog_id:
					if n+i < ivsize:
						continue
					msg(f'\rIncog data for ID {incog_id} found at offset {n+i-ivsize}')
					if not keep_searching:
						sys.exit(0)
			carry = d[len(d)-ivsize:]
			n += bsize
			if not n % mod:
				msg_r(f'\rSearched: {n} bytes')

		msg('')
		os.close(f)
		return True

	def rand2file(self, outfile: str, nbytes: str, *, threads=4, silent=False):
		"""
		write ‘nbytes’ bytes of random data to specified file (dd-style byte specifiers supported)

		Valid specifiers:

		  c  = 1
		  w  = 2
		  b  = 512
		  kB = 1000
		  K  = 1024
		  MB = 1000000
		  M  = 1048576
		  GB = 1000000000
		  G  = 1073741824
		  TB = 1000000000000
		  T  = 1099511627776
		  PB = 1000000000000000
		  P  = 1125899906842624
		  EB = 1000000000000000000
		  E  = 1152921504606846976
		"""
		from threading import Thread
		from queue import Queue
		from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
		from cryptography.hazmat.backends import default_backend

		from ..util2 import parse_bytespec

		def encrypt_worker():
			ctr_init_val = os.urandom(Crypto.aesctr_iv_len)
			c = Cipher(algorithms.AES(key), modes.CTR(ctr_init_val), backend=default_backend())
			encryptor = c.encryptor()
			while True:
				q2.put(encryptor.update(q1.get()))
				q1.task_done()

		def output_worker():
			while True:
				f.write(q2.get())
				q2.task_done()

		nbytes = parse_bytespec(nbytes)
		if self.cfg.outdir:
			outfile = make_full_path(self.cfg.outdir, outfile)

		f = open(outfile, 'wb')

		key = Crypto(self.cfg).get_random(32)
		q1, q2 = (Queue(), Queue())

		for i in range(max(1, threads-2)):
			t = Thread(target=encrypt_worker)
			t.daemon = True
			t.start()

		t = Thread(target=output_worker)
		t.daemon = True
		t.start()

		blk_size = 1024 * 1024
		for i in range(nbytes // blk_size):
			if not i % 4:
				msg_r(f'\rRead: {i * blk_size} bytes')
			q1.put(os.urandom(blk_size))

		if nbytes % blk_size:
			q1.put(os.urandom(nbytes % blk_size))

		q1.join()
		q2.join()
		f.close()

		fsize = os.stat(outfile).st_size
		if fsize != nbytes:
			die(3, f'{fsize}: incorrect random file size (should be {nbytes})')

		if not silent:
			msg(f'\rRead: {nbytes} bytes')
			self.cfg._util.qmsg(f'\r{nbytes} byte{suf(nbytes)} of random data written to file {outfile!r}')

		return True

	def decrypt_keystore(self, wallet_file: str, *, output_hex=False):
		"decrypt the data in a keystore wallet, returning the decrypted data in binary format"
		from ..ui import line_input
		passwd = line_input(self.cfg, 'Enter passphrase: ', echo=self.cfg.echo_passphrase).strip().encode()
		import json
		with open(wallet_file) as fh:
			data = json.loads(fh.read())
		from ..altcoin.util import decrypt_keystore
		ret = decrypt_keystore(data[0]['keystore'], passwd)
		return ret.hex() if output_hex else ret

	def decrypt_geth_keystore(self, wallet_file: str, *, check_addr=True):
		"decrypt the private key in a Geth keystore wallet, returning the decrypted key in hex format"
		from ..ui import line_input
		passwd = line_input(self.cfg, 'Enter passphrase: ', echo=self.cfg.echo_passphrase).strip().encode()
		from ..proto.eth.util import decrypt_geth_keystore
		return decrypt_geth_keystore(self.cfg, wallet_file, passwd, check_addr=check_addr).hex()



#!/usr/bin/env python3
#
# Python 2 code from here: https://github.com/ctz/keccak
# Ported to python3 by the MMGen Project
#
# This is the old, pre-SHA3 version of Keccak used by Ethereum, which is not supported
# by hashlib.sha3

from math import log
from operator import xor
from copy import deepcopy
from functools import reduce

# The Keccak-f round constants.
RoundConstants = [
	0x0000000000000001,   0x0000000000008082,   0x800000000000808A,   0x8000000080008000,
	0x000000000000808B,   0x0000000080000001,   0x8000000080008081,   0x8000000000008009,
	0x000000000000008A,   0x0000000000000088,   0x0000000080008009,   0x000000008000000A,
	0x000000008000808B,   0x800000000000008B,   0x8000000000008089,   0x8000000000008003,
	0x8000000000008002,   0x8000000000000080,   0x000000000000800A,   0x800000008000000A,
	0x8000000080008081,   0x8000000000008080,   0x0000000080000001,   0x8000000080008008
]

RotationConstants = [
	[  0,  1, 62, 28, 27, ],
	[ 36, 44,  6, 55, 20, ],
	[  3, 10, 43, 25, 39, ],
	[ 41, 45, 15, 21,  8, ],
	[ 18,  2, 61, 56, 14, ]
]

Masks = [(1 << i) - 1 for i in range(65)]

def bits2bytes(x):
	return (int(x) + 7) // 8

def rol(value, left, bits):
	"""
	Circularly rotate 'value' to the left,
	treating it as a quantity of the given size in bits.
	"""
	top = value >> (bits - left)
	bot = (value & Masks[bits - left]) << left
	return bot | top

def ror(value, right, bits):
	"""
	Circularly rotate 'value' to the right,
	treating it as a quantity of the given size in bits.
	"""
	top = value >> right
	bot = (value & Masks[right]) << (bits - right)
	return bot | top

def multirate_padding(used_bytes, align_bytes):
	"""
	The Keccak padding function.
	"""
	padlen = align_bytes - used_bytes
	if padlen == 0:
		padlen = align_bytes
	# note: padding done in 'internal bit ordering', wherein LSB is leftmost
	if padlen == 1:
		return [0x81]
	else:
		return [0x01] + ([0x00] * (padlen - 2)) + [0x80]

def keccak_f(state):
	"""
	This is Keccak-f permutation.  It operates on and
	mutates the passed-in KeccakState.  It returns nothing.
	"""
	def round(A, RC):
		W, H = state.W, state.H
		rangeW, rangeH = state.rangeW, state.rangeH
		lanew = state.lanew
		zero = state.zero

		# theta
		C = [reduce(xor, A[x]) for x in rangeW]
		D = [0] * W
		for x in rangeW:
			D[x] = C[(x - 1) % W] ^ rol(C[(x + 1) % W], 1, lanew)
			for y in rangeH:
				A[x][y] ^= D[x]

		# rho and pi
		B = zero()
		for x in rangeW:
			for y in rangeH:
				B[y % W][(2 * x + 3 * y) % H] = rol(A[x][y], RotationConstants[y][x], lanew)

		# chi
		for x in rangeW:
			for y in rangeH:
				A[x][y] = B[x][y] ^ ((~ B[(x + 1) % W][y]) & B[(x + 2) % W][y])

		# iota
		A[0][0] ^= RC

	l = int(log(state.lanew, 2))
	nr = 12 + 2 * l

	for ir in range(nr):
		round(state.s, RoundConstants[ir])

class KeccakState:
	"""
	A keccak state container.

	The state is stored as a 5x5 table of integers.
	"""
	W = 5
	H = 5

	rangeW = range(W)
	rangeH = range(H)

	@staticmethod
	def zero():
		"""
		Returns an zero state table.
		"""
		return [[0] * KeccakState.W for x in KeccakState.rangeH]

	@staticmethod
	def format(st):
		"""
		Formats the given state as hex, in natural byte order.
		"""
		rows = []
		def fmt(x):
			return '%016x' % x
		for y in KeccakState.rangeH:
			row = []
			for x in KeccakState.rangeW:
				row.append(fmt(st[x][y]))
			rows.append(' '.join(row))
		return '\n'.join(rows)

	@staticmethod
	def lane2bytes(s, w):
		"""
		Converts the lane s to a sequence of byte values,
		assuming a lane is w bits.
		"""
		return [(s >> b) & 0xff for b in range(0, w, 8)]

	@staticmethod
	def bytes2lane(bb):
		"""
		Converts a sequence of byte values to a lane.
		"""
		r = 0
		for b in reversed(bb):
			r = r << 8 | b
		return r

	def __init__(self, bitrate, b):
		self.bitrate = bitrate
		self.b = b

		# only byte-aligned
		assert self.bitrate % 8 == 0
		self.bitrate_bytes = bits2bytes(self.bitrate)

		assert self.b % 25 == 0
		self.lanew = self.b // 25

		self.s = KeccakState.zero()

	def __str__(self):
		return KeccakState.format(self.s)

	def absorb(self, bb):
		"""
		Mixes in the given bitrate-length string to the state.
		"""
		assert len(bb) == self.bitrate_bytes

		bb += [0] * bits2bytes(self.b - self.bitrate)
		i = 0

		for y in self.rangeH:
			for x in self.rangeW:
				self.s[x][y] ^= KeccakState.bytes2lane(bb[i:i + 8])
				i += 8

	def squeeze(self):
		"""
		Returns the bitrate-length prefix of the state to be output.
		"""
		return self.get_bytes()[:self.bitrate_bytes]

	def get_bytes(self):
		"""
		Convert whole state to a byte string.
		"""
		out = [0] * bits2bytes(self.b)
		i = 0
		for y in self.rangeH:
			for x in self.rangeW:
				v = KeccakState.lane2bytes(self.s[x][y], self.lanew)
				out[i:i+8] = v
				i += 8
		return out

	def set_bytes(self, bb):
		"""
		Set whole state from byte string, which is assumed
		to be the correct length.
		"""
		i = 0
		for y in self.rangeH:
			for x in self.rangeW:
				self.s[x][y] = KeccakState.bytes2lane(bb[i:i+8])
				i += 8

class KeccakSponge:
	def __init__(self, bitrate, width, padfn, permfn):
		self.state = KeccakState(bitrate, width)
		self.padfn = padfn
		self.permfn = permfn
		self.buffer = []

	def copy(self):
		return deepcopy(self)

	def absorb_block(self, bb):
		assert len(bb) == self.state.bitrate_bytes
		self.state.absorb(bb)
		self.permfn(self.state)

	def absorb(self, s):
		self.buffer += bytes(s)

		while len(self.buffer) >= self.state.bitrate_bytes:
			self.absorb_block(self.buffer[:self.state.bitrate_bytes])
			self.buffer = self.buffer[self.state.bitrate_bytes:]

	def absorb_final(self):
		padded = self.buffer + self.padfn(len(self.buffer), self.state.bitrate_bytes)
		self.absorb_block(padded)
		self.buffer = []

	def squeeze_once(self):
		rc = self.state.squeeze()
		self.permfn(self.state)
		return rc

	def squeeze(self, l):
		Z = self.squeeze_once()
		while len(Z) < l:
			Z += self.squeeze_once()
		return Z[:l]

class KeccakHash:
	"""
	The Keccak hash function, with a hashlib-compatible interface.
	"""
	def __init__(self, bitrate_bits, capacity_bits, output_bits):
		# our in-absorption sponge. this is never given padding
		assert bitrate_bits + capacity_bits in (25, 50, 100, 200, 400, 800, 1600)
		self.sponge = KeccakSponge(bitrate_bits, bitrate_bits + capacity_bits,
								   multirate_padding,
								   keccak_f)

		# hashlib interface members
		assert output_bits % 8 == 0
		self.digest_size = bits2bytes(output_bits)
		self.block_size = bits2bytes(bitrate_bits)

	def __repr__(self):
		inf = (self.sponge.state.bitrate,
			   self.sponge.state.b - self.sponge.state.bitrate,
			   self.digest_size * 8)
		return '<KeccakHash with r=%d, c=%d, image=%d>' % inf

	def copy(self):
		return deepcopy(self)

	def update(self, s):
		self.sponge.absorb(s)

	def digest(self):
		finalised = self.sponge.copy()
		finalised.absorb_final()
		digest = finalised.squeeze(self.digest_size)
		return bytes(digest)

	def hexdigest(self):
		return self.digest().hex()

	@staticmethod
	def preset(bitrate_bits, capacity_bits, output_bits):
		"""
		Returns a factory function for the given bitrate, sponge capacity and output length.
		The function accepts an optional initial input, ala hashlib.
		"""
		def create(initial_input = None):
			h = KeccakHash(bitrate_bits, capacity_bits, output_bits)
			if initial_input is not None:
				h.update(initial_input)
			return h
		return create

# SHA3 parameter presets
keccak_224 = KeccakHash.preset(1152, 448, 224)
keccak_256 = KeccakHash.preset(1088, 512, 256)
keccak_384 = KeccakHash.preset(832, 768, 384)
keccak_512 = KeccakHash.preset(576, 1024, 512)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
devinit: Developer tools initialization for the MMGen suite
"""

devtools_funcs = {
	'pfmt':              lambda *args, **kwargs: devtools_call('pfmt', *args, **kwargs),
	'pmsg':              lambda *args, **kwargs: devtools_call('pmsg', *args, **kwargs),
	'pmsg_r':            lambda *args, **kwargs: devtools_call('pmsg_r', *args, **kwargs),
	'pdie':              lambda *args, **kwargs: devtools_call('pdie', *args, **kwargs),
	'pexit':             lambda *args, **kwargs: devtools_call('pexit', *args, **kwargs),
	'Pmsg':              lambda *args, **kwargs: devtools_call('Pmsg', *args, **kwargs),
	'Pdie':              lambda *args, **kwargs: devtools_call('Pdie', *args, **kwargs),
	'Pexit':             lambda *args, **kwargs: devtools_call('Pexit', *args, **kwargs),
	'print_stack_trace': lambda *args, **kwargs: devtools_call('print_stack_trace', *args, **kwargs),
	'get_diff':          lambda *args, **kwargs: devtools_call('get_diff', *args, **kwargs),
	'print_diff':        lambda *args, **kwargs: devtools_call('print_diff', *args, **kwargs),
	'get_ndiff':         lambda *args, **kwargs: devtools_call('get_ndiff', *args, **kwargs),
	'print_ndiff':       lambda *args, **kwargs: devtools_call('print_ndiff', *args, **kwargs),
}

def devtools_call(funcname, *args, **kwargs):
	from . import devtools
	return getattr(devtools, funcname)(*args, **kwargs)

def MMGenObject_call(methodname, *args, **kwargs):
	from .devtools import MMGenObjectMethods
	return getattr(MMGenObjectMethods, methodname)(*args, **kwargs)

class MMGenObjectDevTools:

	pmsg  = lambda *args, **kwargs: MMGenObject_call('pmsg', *args, **kwargs)
	pdie  = lambda *args, **kwargs: MMGenObject_call('pdie', *args, **kwargs)
	pexit = lambda *args, **kwargs: MMGenObject_call('pexit', *args, **kwargs)
	pfmt  = lambda *args, **kwargs: MMGenObject_call('pfmt', *args, **kwargs)

	# Check that all immutable attrs in ‘valid_attrs’ exist and have been initialized.
	# Expensive, so do only when testing.
	def immutable_attr_init_check(self):

		cls = type(self)

		for attrname in self.valid_attrs:
			# existence check:
			for o in cls.__mro__[:3]: # allow for 2 levels of subclassing
				if attrname in o.__dict__:
					attr = o.__dict__[attrname]
					break
			else:
				from .util import die
				die(4, f'unable to find descriptor {cls.__name__}.{attrname}')

			# initialization check:
			if type(attr).__name__ == 'ImmutableAttr' and attrname not in self.__dict__:
				from .util import die
				die(4, f'attribute {attrname!r} of {cls.__name__} has not been initialized in constructor!')

def init_dev():
	import builtins
	# MMGenObject is added to the namespace by objmethods.py, so we must name the builtin differently
	# to avoid inadvertently adding MMGenObject to the global namespace here:
	setattr(builtins, 'MMGenObjectDevTools', MMGenObjectDevTools)
	for funcname, func in devtools_funcs.items():
		setattr(builtins, funcname, func)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
exception: Exception classes for the MMGen suite
"""

class MMGenError(Exception):

	def __init__(self, errno, strerror, stdout):
		self.mmcode = errno
		self.stdout = stdout
		super().__init__(strerror)

	def __repr__(self):
		return f'{type(self).__name__}({self.mmcode}):\n{self}'

class MMGenSystemExit(MMGenError):

	def __repr__(self):
		return f'{type(self).__name__}({self.mmcode}): {self}'

# 1: no hl, message only
class UserNonConfirmation(Exception):     mmcode = 1
class BadAgeFormat(Exception):            mmcode = 1
class BadFilename(Exception):             mmcode = 1
class BadFileExtension(Exception):        mmcode = 1
class SocketError(Exception):             mmcode = 1
class UserAddressNotInWallet(Exception):  mmcode = 1
class MnemonicError(Exception):           mmcode = 1
class RangeError(Exception):              mmcode = 1
class FileNotFound(Exception):            mmcode = 1
class InvalidPasswdFormat(Exception):     mmcode = 1
class CfgFileParseError(Exception):       mmcode = 1
class UserOptError(Exception):            mmcode = 1
class CmdlineOptError(Exception):         mmcode = 1
class NoLEDSupport(Exception):            mmcode = 1
class MsgFileFailedSID(Exception):        mmcode = 1
class TestSuiteException(Exception):      mmcode = 1
class TestSuiteSpawnedScriptException(Exception): mmcode = 1

# 2: yellow hl, message only
class InvalidContractAddress(Exception):  mmcode = 2
class UnrecognizedTokenSymbol(Exception): mmcode = 2
class TokenNotInBlockchain(Exception):    mmcode = 2
class TokenNotInWallet(Exception):        mmcode = 2
class BadTwComment(Exception):            mmcode = 2
class BadTwLabel(Exception):              mmcode = 2
class BaseConversionError(Exception):     mmcode = 2
class BaseConversionPadError(Exception):  mmcode = 2
class TransactionChainMismatch(Exception):mmcode = 2
class ObjectInitError(Exception):         mmcode = 2
class ClassFlagsError(Exception):         mmcode = 2
class ExtensionModuleError(Exception):    mmcode = 2
class MoneroMMGenTXFileParseError(Exception): mmcode = 2
class AutosignTXError(Exception):         mmcode = 2
class MMGenImportError(Exception):        mmcode = 2
class SwapMemoParseError(Exception):      mmcode = 2
class SwapAssetError(Exception):          mmcode = 2
class SwapCfgValueError(Exception):       mmcode = 2

# 3: yellow hl, 'MMGen Error' + exception + message
class RPCFailure(Exception):              mmcode = 3
class RPCChainMismatch(Exception):        mmcode = 3
class BadTxSizeEstimate(Exception):       mmcode = 3
class MaxInputSizeExceeded(Exception):    mmcode = 3
class MaxFileSizeExceeded(Exception):     mmcode = 3
class MaxFeeExceeded(Exception):          mmcode = 3
class WalletFileError(Exception):         mmcode = 3
class HexadecimalStringError(Exception):  mmcode = 3
class SeedLengthError(Exception):         mmcode = 3
class PrivateKeyError(Exception):         mmcode = 3
class MMGenCalledProcessError(Exception): mmcode = 3
class TestSuiteFatalException(Exception): mmcode = 3

# 4: red hl, 'MMGen Fatal Error' + exception + message
class BadMMGenTxID(Exception):            mmcode = 4
class IllegalWitnessFlagValue(Exception): mmcode = 4
class TxHexParseError(Exception):         mmcode = 4
class TxHexMismatch(Exception):           mmcode = 4
class SubSeedNonceRangeExceeded(Exception): mmcode = 4



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
swap.asset: THORChain swap asset class for the MMGen Wallet suite
"""

from ...asset import SwapAsset

class THORChainSwapAsset(SwapAsset):

	_ad = SwapAsset._ad
	assets_data = {
		'BTC':       _ad('Bitcoin',                 'BTC',  None,        'b',  True),
		'LTC':       _ad('Litecoin',                'LTC',  None,        'l',  True),
		'BCH':       _ad('Bitcoin Cash',            'BCH',  None,        'c',  True),
		'ETH':       _ad('Ethereum',                'ETH',  None,        'e',  True),
		'DOGE':      _ad('Dogecoin',                'DOGE', None,        'd',  False),
		'RUNE':      _ad('Rune (THORChain)',        'RUNE', 'THOR.RUNE', 'r',  True),
		'ETH.AAVE':  _ad('Aave (ETH)',              None,   'ETH.AAVE',  None, True),
		'ETH.DAI':   _ad('MakerDAO USD (ETH)',      None,   'ETH.DAI',   None, True),
		'ETH.DPI':   _ad('DeFi Pulse Index (ETH)',  None,   'ETH.DPI',   None, True),
		'ETH.FOX':   _ad('ShapeShift FOX (ETH)',    None,   'ETH.FOX',   None, True),
		'ETH.GUSD':  _ad('Gemini Dollar (ETH)',     None,   'ETH.GUSD',  None, True),
		'ETH.LINK':  _ad('Chainlink (ETH)',         None,   'ETH.LINK',  None, True),
		'ETH.LUSD':  _ad('Liquity USD (ETH)',       None,   'ETH.LUSD',  None, True),
		'ETH.SNX':   _ad('Synthetix (ETH)',         None,   'ETH.SNX',   None, True),
		'ETH.TGT':   _ad('THORWallet (ETH)',        None,   'ETH.TGT',   None, True),
		'ETH.THOR':  _ad('THORSwap (ETH)',          None,   'ETH.THOR',  None, True),
		'ETH.USDC':  _ad('USDC (ETH)',              None,   'ETH.USDC',  None, True),
		'ETH.USDP':  _ad('Pax Dollar (ETH)',        None,   'ETH.USDP',  None, True),
		'ETH.USDT':  _ad('Tether (ETH)',            None,   'ETH.USDT',  None, True),
		'ETH.vTHOR': _ad('THORSwap Staking (ETH)',  None,   'ETH.vTHOR', None, True),
		'ETH.WBTC':  _ad('Wrapped BTC (ETH)',       None,   'ETH.WBTC',  None, True),
		'ETH.XRUNE': _ad('Thorstarter (ETH)',       None,   'ETH.XRUNE', None, True),
		'ETH.YFI':   _ad('yearn.finance (ETH)',     None,   'ETH.YFI',   None, True),
	}

	evm_contracts = {
		'ETH.AAVE':  '7fc66500c84a76ad7e9c93437bfc5ac33e2ddae9',
		'ETH.DAI':   '6b175474e89094c44da98b954eedeac495271d0f',
		'ETH.DPI':   '1494ca1f11d487c2bbe4543e90080aeba4ba3c2b',
		'ETH.FOX':   'c770eefad204b5180df6a14ee197d99d808ee52d',
		'ETH.GUSD':  '056fd409e1d7a124bd7017459dfea2f387b6d5cd',
		'ETH.LINK':  '514910771af9ca656af840dff83e8264ecf986ca',
		'ETH.LUSD':  '5f98805a4e8be255a32880fdec7f6728c6568ba0',
		'ETH.SNX':   'c011a73ee8576fb46f5e1c5751ca3b9fe0af2a6f',
		'ETH.TGT':   '108a850856db3f85d0269a2693d896b394c80325',
		'ETH.THOR':  'a5f2211b9b8170f694421f2046281775e8468044',
		'ETH.USDC':  'a0b86991c6218b36c1d19d4a2e9eb0ce3606eb48',
		'ETH.USDP':  '8e870d67f660d95d5be530380d0ec0bd388289e1',
		'ETH.USDT':  'dac17f958d2ee523a2206206994597c13d831ec7',
		'ETH.vTHOR': '815c23eca83261b6ec689b60cc4a58b54bc24d8d',
		'ETH.WBTC':  '2260fac5e5542a773aa44fbcfedf7c193bc2c599',
		'ETH.XRUNE': '69fa0fee221ad11012bab0fdb45d444d3d2ce71c',
		'ETH.YFI':   '0bc529c00c6401aef6d220be8c6ea1667f6ad93e',
	}

	unsupported = ('DOGE',)

	blacklisted = {}

	evm_chains = ('ETH', 'AVAX', 'BSC', 'BASE')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.httpd.thornode.rpc: Thornode RPC HTTP server
"""

import re, json
from wsgiref.util import request_uri

from . import ThornodeServer

class ThornodeRPCServer(ThornodeServer):
	port = 18800
	name = 'thornode RPC server'

	def make_response_body(self, method, environ):
		req_str = request_uri(environ)
		if method == 'POST':
			length = int(environ.get('CONTENT_LENGTH', '0'))

		if re.search(r'/bank/balances/(\S+)', req_str):
			res = [
				{'denom': 'foocoin', 'amount': 321321321321},
				{'denom': 'rune',    'amount': 987654321321},
				{'denom': 'barcoin', 'amount': 123123123123}]
		elif m := re.search(r'/auth/accounts/(\S+)', req_str):
			res = {
				'value': {
					'address': m[1],
					'pub_key': 'PubKeySecp256k1{0000}',
					'account_number': '1234',
					'sequence': '333444'}}
		elif m := re.search(r'/tx$', req_str):
			assert method == 'POST'
			txid = environ['wsgi.input'].read(length).decode().removeprefix('hash=0x').upper()
			res = {
				'hash': txid,
				'height': '21298600',
				'index': 2,
				'tx_result': {
					'gas_used': '173222',
					'events': [],
					'codespace': ''
				},
				'tx': 'MHgwMGZvb2Jhcg=='}
		elif m := re.search(r'/check_tx$', req_str):
			assert method == 'POST'
			res = {
				'code': 0,
				'data': '',
				'log': '',
				'info': '',
				'gas_wanted': '-1',
				'gas_used': '53774',
				'events': [],
				'codespace': ''}
		elif m := re.search(r'/broadcast_tx_sync$', req_str):
			assert method == 'POST'
			txhex = environ['wsgi.input'].read(length).decode().removeprefix('tx=0x').upper()
			res = {'code': 0, 'codespace': '', 'data': '', 'log': ''}
			if txhex.startswith('0A540A52'):
				res.update({'hash': '14463C716CF08A814868DB779156BCD85A1DF8EE49E924900A74482E9DEE132D'})
			elif txhex.startswith('0AC1010A'):
				res.update({'hash': '17F9411E48542C0DCA4D40A0DD4A1795DE6D5791A873A27CBBDC1031FE8D1BC5'})
		else:
			raise ValueError(f'‘{req_str}’: malformed query path')

		return json.dumps({'result': res}).encode()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.transaction: Ethereum serialized transaction class
"""

# Adapted from:
#   https://github.com/ethereum/pyethereum/blob/master/ethereum/transactions.py
#   Copyright (c) 2015 Vitalik Buterin, Heiko Hees (MIT License)
#
#   Modified to use libsecp256k1 instead of py_ecc for all EC operations

from enum import Enum

from ....util import cached_property
from ....util2 import get_keccak
from ....protocol import CoinProtocol

from ...secp256k1.secp256k1 import sign_msghash, pubkey_recover, pubkey_gen

from .. import rlp
from ..rlp.sedes import big_endian_int, binary, Binary

secp256k1_ge = CoinProtocol.Secp256k1.secp256k1_group_order
null_address = b'\xff' * 20

keccak_256 = get_keccak()

def sha3(bytes_data):
	return keccak_256(bytes_data).digest()

def mk_contract_addr(sender, nonce):
	return sha3(rlp.encode([sender, nonce]))[12:]

class EthereumTransactionError(Exception):
	pass

class gas_code(Enum):
	GTXCOST        = 21000 # TX BASE GAS COST
	GTXDATAZERO    = 4     # TX DATA ZERO BYTE GAS COST
	GTXDATANONZERO = 68    # TX DATA NON ZERO BYTE GAS COST

class Transaction(rlp.Serializable):

	"""
	A transaction is stored as:
	[nonce, gasprice, startgas, to, value, data, v, r, s]

	nonce is the number of transactions already sent by that account, encoded
	in binary form (eg.  0 -> '', 7 -> '\x07', 1000 -> '\x03\xd8').

	(v,r,s) is the raw Electrum-style signature of the transaction without the
	signature made with the private key corresponding to the sending account,
	with 0 <= v <= 3. From an Electrum-style signature (65 bytes) it is
	possible to extract the public key, and thereby the address, directly.

	A valid transaction is one where:
	(i) the signature is well-formed (ie. 0 <= v <= 3, 0 <= r < P, 0 <= s < N,
	    0 <= r < P - N if v >= 2), and
	(ii) the sending account has enough funds to pay the fee and the value.
	"""

	fields = [
		('nonce',    big_endian_int),
		('gasprice', big_endian_int),
		('startgas', big_endian_int),
		('to',       Binary.fixed_length(20, allow_empty=True)),
		('value',    big_endian_int),
		('data',     binary),
		('v',        big_endian_int),
		('r',        big_endian_int),
		('s',        big_endian_int)]

	def __init__(self, nonce, gasprice, startgas, to, value, data, v=0, r=0, s=0):

		super().__init__(
			nonce,
			gasprice,
			startgas,
			to,
			value,
			data,
			v,
			r,
			s)

	# https://ethereum.stackexchange.com/questions/35481/
	#    deriving-the-v-value-from-an-ecdsa-signature-without-web3j
	#
	# https://ethereum.stackexchange.com/questions/35764/r-s-v-ecdsa-packet-signatures/35770
	#
	#   Basically this code from the bitcoin core libraries define the v value
	#
	#      *recid = (overflow ? 2 : 0) | (secp256k1_fe_is_odd(&r.y) ? 1 : 0);
	#
	#   That gets added to 27 for Ethereum, while 29 and 30 are invalid according to YellowPaper.io
	#
	#   Please note that if an implementation supports EIP-155, instead of using 27 or 28,
	#   it uses CHAIN_ID * 2 + 35. So the value of "v" can also be 37 or 38
	@property
	def recid(self):
		return self.v - 27 if self.v in (27, 28) else self.v - 35 - self.network_id * 2

	@property
	def network_id(self):
		if self.r == 0 and self.s == 0:
			return self.v
		elif self.v in (27, 28):
			return None
		else:
			return ((self.v - 1) // 2) - 17

	def get_sighash(self, network_id):
		if network_id is None:
			return sha3(rlp.encode(unsigned_tx_from_tx(self), UnsignedTransaction))
		else:
			assert 1 <= network_id < 2 ** 63 - 18, f'{network_id}: invalid network ID'
			return sha3(rlp.encode(
				rlp.infer_sedes(self).serialize(self)[:-3]
				+ [network_id, b'', b'']))

	def sign(self, key, network_id=None):
		"""
		Sign transaction with a private key, overwriting any existing signature
		"""
		sig, recid = sign_msghash(self.get_sighash(network_id), bytes.fromhex(key))
		ret = self.copy(
			v = 27 + recid if network_id is None else 35 + recid + network_id * 2,
			r = int.from_bytes(sig[:32], 'big'),
			s = int.from_bytes(sig[32:], 'big'))
		ret._sender = sha3(pubkey_gen(bytes.fromhex(key), 0)[1:])[12:]
		return ret

	@cached_property
	def sender(self):
		"""
		Recover transaction sender from signature
		"""
		if self.r == 0 and self.s == 0:
			return null_address
		# In the yellow paper it is specified that s should be smaller than secp256k1_ge (eq.205)
		elif self.r >= secp256k1_ge or self.s >= secp256k1_ge or self.r == 0 or self.s == 0:
			raise EthereumTransactionError('Invalid signature (r or s) values!')
		return sha3(pubkey_recover(
			self.get_sighash(self.network_id),
			self.r.to_bytes(32, 'big') + self.s.to_bytes(32, 'big'),
			self.recid,
			False)[1:])[12:]

	@property
	def hash(self):
		return sha3(rlp.encode(self))

	def to_dict(self):
		d = {}
		for name, _ in self.__class__._meta.fields:
			d[name] = getattr(self, name)
			if name in ('to', 'data'):
				d[name] = '0x' + d[name].hex()
		d['sender'] = '0x' + self.sender.hex()
		d['hash']   = '0x' + self.hash.hex()
		return d

	@property
	def intrinsic_gas_used(self):
		num_zero_bytes = self.data.count(0)
		num_non_zero_bytes = len(self.data) - num_zero_bytes
		return (
			gas_code.GTXCOST
			# + (0 if self.to else gas_code.CREATE[3])
			+ gas_code.GTXDATAZERO * num_zero_bytes
			+ gas_code.GTXDATANONZERO * num_non_zero_bytes)

	@property
	def creates(self):
		'returns the address of the contract created by this transaction'
		if self.to in (b'', bytes(20)):
			return mk_contract_addr(self.sender, self.nonce)

	def __eq__(self, other):
		return isinstance(other, self.__class__) and self.hash == other.hash

	def __lt__(self, other):
		return isinstance(other, self.__class__) and self.hash < other.hash

	def __hash__(self):
		return int.from_bytes(self.hash, 'big')

	def __ne__(self, other):
		return not self.__eq__(other)

	def __repr__(self):
		return '<Transaction_(%s)>' % self.hash.hex()[:8]

	def __structlog__(self):
		return self.hash.hex()

class UnsignedTransaction(rlp.Serializable):
	fields = [(field, sedes) for field, sedes in Transaction._meta.fields if field not in 'vrs']

def unsigned_tx_from_tx(tx):
	return UnsignedTransaction(
		nonce    = tx.nonce,
		gasprice = tx.gasprice,
		startgas = tx.startgas,
		to       = tx.to,
		value    = tx.value,
		data     = tx.data)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
key: MMGen public and private key objects
"""

from .objmethods import HiliteStr, InitErrors, MMGenObject
from .obj import ImmutableAttr, get_obj

class WifKey(HiliteStr, InitErrors):
	"""
	Initialize a WIF key, checking its well-formedness.
	The numeric validity of the private key it encodes is not checked.
	"""
	width = 53
	color = 'blue'
	def __new__(cls, proto, wif):
		if isinstance(wif, cls):
			return wif
		try:
			assert wif.isascii() and wif.isalnum(), 'not an ASCII alphanumeric string'
			proto.decode_wif(wif) # raises exception on error
			return str.__new__(cls, wif)
		except Exception as e:
			return cls.init_fail(e, wif)

def is_wif(proto, s):
	return get_obj(WifKey, proto=proto, wif=s, silent=True, return_bool=True)

class PubKey(bytes, InitErrors, MMGenObject): # TODO: add some real checks

	def __new__(cls, s, compressed):
		try:
			assert isinstance(s, bytes)
			me = bytes.__new__(cls, s)
			me.compressed = compressed
			return me
		except Exception as e:
			return cls.init_fail(e, s)

class PrivKey(bytes, InitErrors, MMGenObject):
	"""
	Input:   a) raw, non-preprocessed bytes; or b) WIF key.
	Output:  preprocessed key bytes, plus WIF key in 'wif' attribute
	For coins without a WIF format, 'wif' contains the preprocessed hex.
	The numeric validity of the resulting key is always checked.
	"""
	color = 'red'
	width = 32
	trunc_ok = False

	compressed = ImmutableAttr(bool, typeconv=False)
	wif        = ImmutableAttr(WifKey, typeconv=False)

	# initialize with (priv_bin, compressed), WIF or self
	def __new__(cls, proto, s=None, *, compressed=None, wif=None, pubkey_type=None):
		if isinstance(s, cls):
			return s
		if wif:
			try:
				assert s is None, "'wif' and key hex args are mutually exclusive"
				assert wif.isascii() and wif.isalnum(), 'not an ASCII alphanumeric string'
				k = proto.decode_wif(wif) # raises exception on error
				me = bytes.__new__(cls, k.sec)
				me.compressed = k.compressed
				me.pubkey_type = k.pubkey_type
				me.wif = str.__new__(WifKey, wif) # check has been done
				me.orig_bytes = None
				if k.sec != proto.preprocess_key(k.sec, k.pubkey_type):
					from .util import die
					die('PrivateKeyError',
						f'{proto.cls_name} WIF key {me.wif!r} encodes private key with invalid value {me}')
				me.proto = proto
				return me
			except Exception as e:
				return cls.init_fail(e, s, objname=f'{proto.coin} WIF key')
		else:
			try:
				assert s, 'private key bytes data missing'
				assert isinstance(s, bytes), 'input is not bytes'
				assert pubkey_type is not None, "'pubkey_type' arg missing"
				assert len(s) == cls.width, f'key length must be {cls.width} bytes'
				if pubkey_type == 'password': # skip WIF creation and pre-processing for passwds
					me = bytes.__new__(cls, s)
				else:
					assert compressed is not None, "'compressed' arg missing"
					assert type(compressed) is bool, (
						f"'compressed' must be of type bool, not {type(compressed).__name__}")
					me = bytes.__new__(cls, proto.preprocess_key(s, pubkey_type))
					me.wif = WifKey(proto, proto.encode_wif(me, pubkey_type, compressed=compressed))
					me.compressed = compressed
				me.pubkey_type = pubkey_type
				me.orig_bytes = s # save the non-preprocessed key
				me.proto = proto
				return me
			except Exception as e:
				return cls.init_fail(e, s)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.incog_base: incognito wallet base class
"""

from ..seed import Seed
from ..util import msg, make_chksum_8, die
from .enc import wallet

class wallet(wallet):

	_msg = {
		'check_incog_id': """
  Check the generated Incog ID above against your records.  If it doesn't
  match, then your incognito data is incorrect or corrupted.
	""",
		'record_incog_id': """
  Make a record of the Incog ID but keep it secret.  You will use it to
  identify your incog wallet data in the future.
	""",
		'decrypt_params': " {} hash preset"
	}

	def _make_iv_chksum(self, s):
		from hashlib import sha256
		return sha256(s).hexdigest()[:8].upper()

	def _get_incog_data_len(self, seed_len):
		return (
			self.crypto.aesctr_iv_len
			+ self.crypto.salt_len
			+ (0 if self.cfg.old_incog_fmt else self.crypto.hincog_chk_len)
			+ seed_len//8)

	def _incog_data_size_chk(self):
		# valid sizes: 56, 64, 72
		dlen = len(self.fmt_data)
		seed_len = self.cfg.seed_len or Seed.dfl_len
		valid_dlen = self._get_incog_data_len(seed_len)
		if dlen == valid_dlen:
			return True
		else:
			if self.cfg.old_incog_fmt:
				msg('WARNING: old-style incognito format requested.  Are you sure this is correct?')
			msg(f'Invalid incognito data size ({dlen} bytes) for this seed length ({seed_len} bits)')
			msg(f'Valid data size for this seed length: {valid_dlen} bytes')
			for sl in Seed.lens:
				if dlen == self._get_incog_data_len(sl):
					die(1, f'Valid seed length for this data size: {sl} bits')
			msg(f'This data size ({dlen} bytes) is invalid for all available seed lengths')
			return False

	def _encrypt (self):
		self._get_first_pw_and_hp_and_encrypt_seed()
		if self.cfg.old_incog_fmt:
			die(1, 'Writing old-format incognito wallets is unsupported')
		d = self.ssdata
		crypto = self.crypto

		d.iv = crypto.get_random(crypto.aesctr_iv_len)
		d.iv_id = self._make_iv_chksum(d.iv)
		msg(f'New Incog Wallet ID: {d.iv_id}')
		self.cfg._util.qmsg('Make a record of this value')
		self.cfg._util.vmsg('\n  ' + self.msg['record_incog_id'].strip()+'\n')

		d.salt = crypto.get_random(crypto.salt_len)
		seed_key = crypto.make_key(
			passwd      = d.passwd,
			salt        = d.salt,
			hash_preset = d.hash_preset,
			desc        = 'incog wallet key')

		from hashlib import sha256
		chk = sha256(self.seed.data).digest()[:8]
		d.enc_seed = crypto.encrypt_seed(
			data = chk + self.seed.data,
			key  = seed_key)

		# IV is used BOTH to initialize counter and to salt password!
		d.wrapper_key = crypto.make_key(
			passwd      = d.passwd,
			salt        = d.iv,
			hash_preset = d.hash_preset,
			desc        = 'incog wrapper key')

		d.key_id = make_chksum_8(d.wrapper_key)
		self.cfg._util.vmsg(f'Key ID: {d.key_id}')

		d.target_data_len = self._get_incog_data_len(self.seed.bitlen)

	def _format(self):
		d = self.ssdata
		self.fmt_data = d.iv + self.crypto.encrypt_data(
			data = d.salt + d.enc_seed,
			key  = d.wrapper_key,
			iv   = d.iv,
			desc = self.desc)

	def _filename(self):
		s = self.seed
		d = self.ssdata
		return '{}-{}-{}[{},{}].{}'.format(
			s.fn_stem,
			d.key_id,
			d.iv_id,
			s.bitlen,
			d.hash_preset,
			self.ext)

	def _deformat(self):

		if not self._incog_data_size_chk():
			return False

		d = self.ssdata
		d.iv             = self.fmt_data[0:self.crypto.aesctr_iv_len]
		d.incog_id       = self._make_iv_chksum(d.iv)
		d.enc_incog_data = self.fmt_data[self.crypto.aesctr_iv_len:]
		msg(f'Incog Wallet ID: {d.incog_id}')
		self.cfg._util.qmsg('Check this value against your records')
		self.cfg._util.vmsg('\n  ' + self.msg['check_incog_id'].strip()+'\n')

		return True

	def _verify_seed_newfmt(self, data):
		chk, seed = data[:8], data[8:]
		from hashlib import sha256
		if sha256(seed).digest()[:8] == chk:
			self.cfg._util.qmsg('Passphrase{} are correct'.format(self.msg['decrypt_params'].format('and')))
			return seed
		else:
			msg('Incorrect passphrase{}'.format(self.msg['decrypt_params'].format('or')))
			return False

	def _verify_seed_oldfmt(self, seed):
		prompt = f'Seed ID: {make_chksum_8(seed)}.  Is the Seed ID correct?'
		from ..ui import keypress_confirm
		if keypress_confirm(self.cfg, prompt, default_yes=True):
			return seed
		else:
			return False

	def _decrypt(self):
		d = self.ssdata
		self._get_hash_preset(add_desc=d.incog_id)
		d.passwd = self._get_passphrase(add_desc=d.incog_id)
		crypto = self.crypto

		# IV is used BOTH to initialize counter and to salt password!
		wrapper_key = crypto.make_key(
			passwd      = d.passwd,
			salt        = d.iv,
			hash_preset = d.hash_preset,
			desc        = 'wrapper key')

		dd = crypto.decrypt_data(
			enc_data = d.enc_incog_data,
			key      = wrapper_key,
			iv       = d.iv,
			desc     = 'incog data')

		d.salt     = dd[0:crypto.salt_len]
		d.enc_seed = dd[crypto.salt_len:]

		seed_key = crypto.make_key(
			passwd      = d.passwd,
			salt        = d.salt,
			hash_preset = d.hash_preset,
			desc        = 'main key')

		self.cfg._util.qmsg(f'Key ID: {make_chksum_8(seed_key)}')

		verify_seed_func = getattr(self, '_verify_seed_'+ ('oldfmt' if self.cfg.old_incog_fmt else 'newfmt'))

		seed = verify_seed_func(
			crypto.decrypt_seed(
				enc_seed = d.enc_seed,
				key      = seed_key,
				seed_id  = '',
				key_id   = ''))

		if seed:
			self.seed = Seed(self.cfg, seed_bin=seed)
			msg(f'Seed ID: {self.seed.sid}')
			return True
		else:
			return False



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
rpc.remote: remote RPC client class for the MMGen Project
"""

class RemoteRPCClient:

	is_remote = True

	def __init__(self, cfg, proto):
		self.cfg = cfg
		self.proto = proto

	def get_block_from_minconf(self, minconf):
		return None



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
rpc.backends.base: base RPC backend class for the MMGen Project
"""

class base:

	def __init__(self, caller):
		self.cfg            = caller.cfg
		self.host           = caller.host
		self.port           = caller.port
		self.proxy          = caller.proxy
		self.host_url       = caller.host_url
		self.timeout        = caller.timeout
		self.http_hdrs      = caller.http_hdrs
		self.name           = type(self).__name__
		self.caller         = caller



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.bch.cashaddr: Bitcoin Cash cashaddr implementation for the MMGen Project
"""

# Specification: https://upgradespecs.bitcoincashnode.org/cashaddr

import re
from collections import namedtuple

b32_matrix = """
	q  p  z  r  y  9  x  8
	g  f  2  t  v  d  w  0
	s  3  j  n  5  4  k  h
	c  e  6  m  u  a  7  l
"""

b32a = re.sub(r'\s', '', b32_matrix)

cashaddr_addr_types = {
	'p2pkh':        0,
	'p2sh':         1,
	'token_pubkey': 2,
	'token_script': 3,
	'unknown':      15,
}
addr_types_rev = {v:k for k, v in cashaddr_addr_types.items()}

data_sizes = (160, 192, 224, 256, 320, 384, 448, 512)

def PolyMod(v):
	c = 1
	for d in v:
		c0 = c >> 35
		c = ((c & 0x07ffffffff) << 5) ^ d
		if (c0 & 1): c ^= 0x98f2bc8e61
		if (c0 & 2): c ^= 0x79b76d99e2
		if (c0 & 4): c ^= 0xf33e5fb3c4
		if (c0 & 8): c ^= 0xae2eabe2a8
		if (c0 & 16): c ^= 0x1e4f43e470
	return c ^ 1

def parse_ver_byte(ver):
	assert not (ver >> 7), 'invalid version byte: most-significant bit must be zero'
	t = namedtuple('parsed_version_byte', ['addr_type', 'bitlen'])
	return t(addr_types_rev[ver >> 3], data_sizes[ver & 7])

def make_ver_byte(addr_type, bitlen):
	assert addr_type in addr_types_rev, f'{addr_type}: invalid addr type'
	return (addr_type << 3) | data_sizes.index(bitlen)

def bin2vec(data_bin):
	assert not len(data_bin) % 5, f'{len(data_bin)}: data length not a multiple of 5'
	return [int(data_bin[i*5:(i*5)+5], 2) for i in range(len(data_bin) // 5)]

def make_polymod_vec(pfx, payload_vec):
	return ([ord(c) & 31  for c in pfx] + [0] + payload_vec)

def cashaddr_parse_addr(addr):
	t = namedtuple('parsed_cashaddr', ['pfx', 'payload'])
	return t(*addr.split(':', 1))

def cashaddr_encode_addr(addr_type, size, pfx, data):
	t = namedtuple('encoded_cashaddr', ['addr', 'pfx', 'payload'])
	payload_bin = (
		'{:08b}'.format(make_ver_byte(addr_type, size * 8)) +
		'{:0{w}b}'.format(int.from_bytes(data, 'big'), w=len(data) * 8)
	)
	payload_vec = bin2vec(payload_bin + '0' * (-len(payload_bin) % 5))
	chksum_vec = bin2vec('{:040b}'.format(PolyMod(make_polymod_vec(pfx, payload_vec + [0] * 8))))
	payload = ''.join(b32a[i] for i in payload_vec + chksum_vec)
	return t(f'{pfx}:{payload}', pfx, payload)

def cashaddr_decode_addr(addr):
	t = namedtuple('decoded_cashaddr', ['pfx', 'payload', 'addr_type', 'bytes', 'chksum'])
	a = cashaddr_parse_addr(addr.lower())
	data_bin = ''.join(f'{b32a.index(c):05b}' for c in a.payload)
	vi = parse_ver_byte(int(data_bin[:8], 2))
	assert len(data_bin) >= vi.bitlen + 48, 'cashaddr data length too short!'
	data    = int(data_bin[8:8+vi.bitlen], 2).to_bytes(vi.bitlen // 8, 'big')
	chksum  = int(data_bin[-40:], 2).to_bytes(5, 'big')
	pad_bin = data_bin[8+vi.bitlen:-40]
	assert not pad_bin or pad_bin in '0000', f'{pad_bin}: invalid cashaddr data'
	if chksum_chk := PolyMod(make_polymod_vec(a.pfx, [b32a.index(c) for c in a.payload])) != 0:
		raise ValueError(
			'checksum check failed\n'
			f'  address:  {addr}\n'
			f'  result:   0x{chksum_chk:x}\n'
			f'  checksum: 0x{chksum.hex()}')
	return t(a.pfx, a.payload, vi.addr_type, data, chksum)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.ethswap: Ethereum swap tests for the cmdtest.py test suite
"""

from subprocess import run, PIPE, DEVNULL

from mmgen.util import msg_r, rmsg, die
from mmgen.protocol import init_proto
from mmgen.fileutil import get_data_from_file

from ..include.common import imsg, chk_equal

from .include.common import dfl_sid, eth_inbound_addr, thorchain_router_addr_file
from .include.proxy import TestProxy
from .httpd.thornode.swap import ThornodeSwapServer

from .regtest import CmdTestRegtest
from .swap import CmdTestSwapMethods, create_cross_methods
from .ethdev import CmdTestEthdev

class CmdTestEthSwapMethods:

	async def token_deploy_a(self):
		return await self._token_deploy_math(num=1)

	async def token_deploy_b(self):
		return await self._token_deploy_owned(num=1)

	async def token_deploy_c(self):
		return await self._token_deploy_token(num=1)

	def token_compile_router(self):

		if not self.using_solc:
			bin_fn = 'test/ref/ethereum/bin/THORChain_Router.bin'
			imsg(f'Using precompiled contract data ‘{bin_fn}’')
			import shutil
			shutil.copy(bin_fn, self.tmpdir)
			return 'skip'

		imsg("Compiling THORChain router contract")
		self.spawn(msg_only=True)
		cmd = [
			'solc',
			'--evm-version=constantinople',
			'--overwrite',
			f'--output-dir={self.tmpdir}',
			'--bin',
			'test/ref/ethereum/THORChain_Router.sol']
		imsg('Executing: {}'.format(' '.join(cmd)))
		cp = run(cmd, stdout=DEVNULL, stderr=PIPE)
		if cp.returncode != 0:
			rmsg('solc failed with the following output:')
			die(2, cp.stderr.decode())
		imsg('THORChain router contract compiled')
		return 'ok'

	async def token_deploy_router(self):
		return await self._token_deploy(
			key = 'thorchain_router',
			gas = 1_000_000,
			fn  = f'{self.tmpdir}/THORChain_Router.bin')

	def token_fund_user(self):
		return self._token_transfer_ops(
			op          = 'fund_user',
			mm_idxs     = [1, 2, 12],
			token_addr  = 'token_addr1',
			amt         = self.token_fund_amt)

	def token_addrgen(self):
		return self._token_addrgen(mm_idxs=[1], naddrs=12)

	def token_addrimport(self):
		return self._token_addrimport('token_addr1', '1-12', expect='12/12')

	def token_addrimport_inbound(self):
		token_addr = self.read_from_tmpfile('token_addr1').strip()
		return self.spawn(
			'mmgen-addrimport',
			['--quiet', '--regtest=1', f'--token-addr={token_addr}', f'--address={eth_inbound_addr}'])

	def token_bal1(self):
		return self._token_bal_check(pat=rf'{dfl_sid}:E:1\s+{self.token_fund_amt}\s')

	def token_bal2(self):
		return self._token_bal_check(pat=rf'{eth_inbound_addr}\s+\S+\s+87.654321\s')

	async def _check_token_swaptx_memo(self, chk):
		from mmgen.proto.eth.contract import Contract
		self.spawn(msg_only=True)
		addr = get_data_from_file(self.cfg, thorchain_router_addr_file, quiet=True).strip()
		c = Contract(self.cfg, self.proto, addr, rpc=await self.rpc)
		res = (await c.do_call('saved_memo()'))[2:]
		memo_len = int(res[64:128], 16)
		chk_equal(bytes.fromhex(res[128:128+(2*memo_len)]).decode(), chk)
		imsg(f'saved_memo: {chk}')
		return 'ok'

	def _swaptxsend_eth_proxy(self, *, add_opts=[], test=False):
		t = self._swaptxsend(
			add_opts = ['--tx-proxy=eth'] + (['--test'] if test else []) + add_opts,
			spawn_only = True)
		t.expect('view: ', 'y')
		t.expect('continue: ', '\n') # exit swap quote
		t.expect('(y/N): ', '\n')    # add comment
		if not test:
			t.expect('to confirm: ', 'YES\n')
		return t

class CmdTestEthSwap(CmdTestSwapMethods, CmdTestRegtest):
	'Ethereum swap operations'

	bdb_wallet = True
	tmpdir_nums = [47]
	networks = ('btc',)
	passthru_opts = ('coin', 'rpc_backend', 'eth_daemon_id')
	cross_group = 'ethswap_eth'
	cross_coin = 'eth'

	cmd_group_in = (
		('setup',                   'regtest (Bob and Alice) mode setup'),
		('eth_setup',               'Ethereum devnet setup'),
		('subgroup.init',           []),
		('subgroup.fund',           ['init']),
		('subgroup.eth_init',       []),
		('subgroup.eth_fund',       ['eth_init']),
		('subgroup.swap',           ['fund', 'eth_fund']),
		('subgroup.eth_swap',       ['fund', 'eth_fund']),
		('subgroup.token_init',     ['eth_fund']),
		('subgroup.token_swap',     ['fund', 'token_init']),
		('subgroup.eth_token_swap', ['fund', 'token_init']),
		('swap_server_stop',        'stopping the Thornode server'),
		('eth_stop',                'stopping the Ethereum daemon'),
		('stop',                    'stopping the regtest daemon'),
	)
	cmd_subgroups = {
	'init': (
		'creating Bob’s MMGen wallet and tracking wallet',
		('walletconv_bob', 'wallet creation (Bob)'),
		('addrgen_bob',    'address generation (Bob)'),
		('addrimport_bob', 'importing Bob’s addresses'),
	),
	'fund': (
		'funding Bob’s wallet',
		('bob_import_miner_addr', 'importing miner’s coinbase addr into Bob’s wallet'),
		('fund_bob',              'funding Bob’s wallet'),
		('generate',              'mining a block'),
		('bob_bal1',              'Bob’s balance'),
	),
	'eth_init': (
		'initializing the ETH tracking wallet',
		('eth_addrgen',                 ''),
		('eth_addrimport',              ''),
		('eth_addrimport_devaddr',      ''),
		('eth_addrimport_reth_devaddr', ''),
		('eth_fund_devaddr',            ''),
		('eth_del_reth_devaddr',        ''),
	),
	'eth_fund': (
		'funding the ETH tracking wallet',
		('eth_fund_mmgen_addr1',  ''),
		('eth_fund_mmgen_addr1b', ''),
		('eth_fund_mmgen_addr2',  ''),
		('eth_bal1',              ''),
	),
	'token_init': (
		'deploying tokens and initializing the ETH token tracking wallet',
		('eth_token_compile1',           ''),
		('eth_token_deploy_a',           ''),
		('eth_token_deploy_b',           ''),
		('eth_token_deploy_c',           ''),
		('eth_token_compile_router',     ''),
		('eth_token_deploy_router',      ''),
		('eth_token_fund_user',          ''),
		('eth_token_addrgen',            ''),
		('eth_token_addrimport',         ''),
		('eth_token_addrimport_inbound', ''),
		('eth_token_bal1',               ''),
	),
	'token_swap': (
		'token swap operations (BTC -> MM1)',
		('swaptxcreate3', 'creating a BTC->MM1 swap transaction'),
		('swaptxsign3',   'signing the swap transaction'),
		('swaptxsend3',   'sending the swap transaction'),
	),
	'swap': (
		'swap operations (BTC -> ETH)',
		('swaptxcreate1', 'creating a BTC->ETH swap transaction'),
		('swaptxcreate2', 'creating a BTC->ETH swap transaction (used account)'),
		('swaptxsign1',   'signing the swap transaction'),
		('swaptxsend1',   'sending the swap transaction'),
		('swaptxbump1',   'bumping the swap transaction'),
		('swaptxsign2',   'signing the bump transaction'),
		('swaptxsend2',   'sending the bump transaction'),
		('generate',      'generating a block'),
		('bob_bal2',      'Bob’s balance'),
		('swaptxdo1',     'creating, signing and sending a swap transaction'),
		('generate',      'generating a block'),
		('bob_bal3',      'Bob’s balance'),
	),
	'eth_swap': (
		'swap operations (ETH -> BTC)',
		('eth_swaptxcreate1', ''),
		('eth_swaptxcreate2', ''),
		('eth_swaptxsign1',   ''),
		('eth_swaptxsend1',   ''),
		('eth_swaptxstatus1', ''),
		('eth_bal2',          ''),
	),
	'eth_token_swap': (
		'swap operations (ETH -> ERC20, ERC20 -> BTC, ERC20 -> ETH)',
		# ETH -> MM1
		('eth_swaptxcreate3a',         ''),
		('eth_swaptxcreate3b',         ''),
		('eth_swaptxsign3',            ''),
		('eth_swaptxsend3',            ''),
		('eth_swaptxmemo3',            ''),
		# MM1 -> BTC
		('eth_swaptxcreate4',          ''),
		('eth_swaptxsign4',            ''),
		('eth_swaptxsend4',            ''),
		('eth_swaptxmemo4',            ''),
		('eth_swaptxstatus4',          ''),
		('eth_swaptxreceipt4',         ''),
		('eth_token_bal2',             ''),
		# MM1 -> ETH
		('eth_swaptxcreate5a',         ''),
		('eth_swaptxcreate5b',         ''),
		('eth_swaptxsign5',            ''),
		('eth_etherscan_server_start', ''),
		('eth_swaptxsend5_test',       ''),
		('eth_swaptxsend5a',           ''),
		('eth_swaptxsend5b',           ''),
		('eth_swaptxsend5',            ''),
		('eth_etherscan_server_stop',  ''),
	),
	}

	exec(create_cross_methods(cross_coin, cross_group, cmd_group_in, cmd_subgroups))

	def __init__(self, cfg, trunner, cfgs, spawn):

		super().__init__(cfg, trunner, cfgs, spawn)

		if not trunner:
			return

		globals()[self.cross_group] = self.create_cross_runner(
			trunner,
			add_cfg = {'eth_daemon_id': trunner.cfg.eth_daemon_id})

		self.swap_server = ThornodeSwapServer()
		self.swap_server.start()

		TestProxy(self, cfg)

	def swaptxcreate1(self):
		t = self._swaptxcreate(['BTC', '8.765', 'ETH'])
		t.expect('OK? (Y/n): ', 'y')
		t.expect(':E:2')
		t.expect('OK? (Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t)

	def swaptxcreate2(self):
		t = self._swaptxcreate(['BTC', '8.765', 'ETH', f'{dfl_sid}:E:1'])
		t.expect('OK? (Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t)

	def swaptxsign1(self):
		return self._swaptxsign()

	def swaptxsend1(self):
		return self._swaptxsend()

	def swaptxsend2(self):
		return self._swaptxsend(add_opts=[f'--proxy=localhost:{TestProxy.port}'])

	swaptxsign3 = swaptxsign2 = swaptxsign1
	swaptxsend3 = swaptxsend1

	def swaptxbump1(self): # create one-output TX back to self to rescue funds
		return self._swaptxbump('40s', output_args=[f'{dfl_sid}:B:1'])

	def swaptxdo1(self):
		return self._swaptxcreate_ui_common(
			self._swaptxcreate(
				['BTC', '0.223344', f'{dfl_sid}:B:3', 'ETH', f'{dfl_sid}:E:2'],
				action = 'txdo'),
			sign_and_send = True,
			file_desc = 'Sent transaction')

	def bob_bal2(self):
		return self._user_bal_cli('bob', chk='499.9999252')

	def bob_bal3(self):
		return self._user_bal_cli('bob', chk='499.77656902')

	def swaptxcreate3(self):
		t = self._swaptxcreate(['BTC', '0.87654321', 'ETH.MM1', f'{dfl_sid}:E:5'])
		t.expect('OK? (Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t)

	def swap_server_stop(self):
		return self._thornode_server_stop()

class CmdTestEthSwapEth(CmdTestEthSwapMethods, CmdTestSwapMethods, CmdTestEthdev):
	'Ethereum swap operations - Ethereum wallet'

	networks = ('eth',)
	tmpdir_nums = [48]
	fund_amt = '123.456'
	token_fund_amt = 1000
	is_helper = True

	bals = lambda self, k: {
		'swap1': [('98831F3A:E:1', '123.456')],
		'swap2': [('98831F3A:E:1', {'geth': '114.690978056', 'reth': '114.69097664'}[self.daemon.id])],
	}[k]

	cmd_group_in = CmdTestEthdev.cmd_group_in + (
		# eth_fund:
		('fund_mmgen_addr1',         'funding user address :1)'),
		('fund_mmgen_addr1b',        'funding user address :3)'),
		('fund_mmgen_addr2',         'funding user address :11)'),
		('bal1',                     'the ETH balance'),
		# eth_swap:
		('swaptxcreate1',            'creating an ETH->BTC swap transaction'),
		('swaptxcreate2',            'creating an ETH->BTC swap transaction (spec address, trade limit)'),
		('swaptxsign1',              'signing the transaction'),
		('swaptxsend1',              'sending the transaction'),
		('swaptxstatus1',            'getting the transaction status (with --verbose)'),
		('bal2',                     'the ETH balance'),
		# token_init:
		('token_compile1',           'compiling ERC20 token #1'),
		('token_deploy_a',           'deploying ERC20 token MM1 (SafeMath)'),
		('token_deploy_b',           'deploying ERC20 token MM1 (Owned)'),
		('token_deploy_c',           'deploying ERC20 token MM1 (Token)'),
		('token_compile_router',     'compiling THORChain router contract'),
		('token_deploy_router',      'deploying THORChain router contract'),
		('token_fund_user',          'transferring token funds from dev to user'),
		('token_addrgen',            'generating token addresses'),
		('token_addrimport',         'importing token addresses using token address (MM1)'),
		('token_addrimport_inbound', 'importing THORNode inbound token address'),
		('token_bal1',               'the token balance'),

		# eth_token_swap:
		# ETH -> MM1
		('swaptxcreate3a',         'creating an ETH->MM1 swap transaction'),
		('swaptxcreate3b',         'creating an ETH->MM1 swap transaction (specific address)'),
		('swaptxsign3',            'signing the transaction'),
		('swaptxsend3',            'sending the transaction'),
		('swaptxmemo3',            'the memo of the sent transaction'),

		# MM1 -> BTC
		('swaptxcreate4',          'creating an MM1->BTC swap transaction'),
		('swaptxsign4',            'signing the transaction'),
		('swaptxsend4',            'sending the transaction'),
		('swaptxmemo4',            'checking the memo'),
		('swaptxstatus4',          'getting the transaction status'),
		('swaptxreceipt4',         'getting the transaction receipt'),
		('token_bal2',             'the token balance'),

		# MM1 -> ETH
		('swaptxcreate5a',         'creating an MM1->ETH swap transaction'),
		('swaptxcreate5b',         'creating an MM1->ETH swap transaction (specific address)'),
		('swaptxsign5',            'signing the transaction'),
		('etherscan_server_start', 'starting the Etherscan server'),
		('swaptxsend5_test',       'testing the transaction via Etherscan'),
		('swaptxsend5a',           'sending the transaction via Etherscan (p1)'),
		('swaptxsend5b',           'sending the transaction via Etherscan (p2)'),
		('swaptxsend5',            'sending the transaction via Etherscan (complete)'),
		('etherscan_server_stop',  'stopping the Etherscan server'),
	)

	def fund_mmgen_addr1b(self):
		return self._fund_mmgen_addr(arg=f'{dfl_sid}:E:3,0.001')

	def swaptxcreate1(self):
		t = self._swaptxcreate(['ETH', '8.765', 'BTC'])
		t.expect('OK? (Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t)

	def swaptxcreate2(self):
		return self._swaptxcreate_ui_common(
			self._swaptxcreate(
				['ETH', '8.765', 'BTC', f'{dfl_sid}:B:4'],
				add_opts = ['--trade-limit=3%' ,'--stream-interval=7']),
			expect = ':2019e4/7/0')

	def swaptxcreate3a(self):
		t = self._swaptxcreate(['ETH', '0.7654321', 'ETH.MM1'], add_opts=['--gas=fallback'])
		t.expect(f'{dfl_sid}:E:4') # check that correct unused address was found
		t.expect('(Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t)

	def swaptxcreate3b(self):
		t = self._swaptxcreate(['ETH', '8.765', 'ETH.MM1', f'{dfl_sid}:E:5'], add_opts=['--gas=auto'])
		return self._swaptxcreate_ui_common(t)

	async def swaptxmemo3(self):
		self.spawn(msg_only=True)
		import json
		fn = self.get_file_with_ext('sigtx')
		tx = json.loads(get_data_from_file(self.cfg, fn, quiet=True).strip())
		txid = tx['MMGenTransaction']['coin_txid']
		chk = '=:ETH.MM1:0x48596c861c970eb4ca72c5082ff7fecd8ee5be9d:0/3/0' # E:5
		imsg(f'TxID: {txid}\nmemo: {chk}')
		res = await (await self.rpc).call('eth_getTransactionByHash', '0x' + txid)
		chk_equal(bytes.fromhex(res['input'].removeprefix('0x')).decode(), chk)
		return 'ok'

	def swaptxcreate4(self):
		t = self._swaptxcreate(['ETH.MM1', '87.654321', 'BTC', f'{dfl_sid}:C:2'], add_opts=['--gas=auto'])
		return self._swaptxcreate_ui_common(t)

	def swaptxcreate5a(self):
		t = self._swaptxcreate(
			['ETH.MM1', '98.7654321', 'ETH'],
			add_opts = ['--gas=58000', '--router-gas=500000'])
		t.expect(f'{dfl_sid}:E:13') # check that correct unused address was found
		t.expect('(Y/n): ', 'y')
		return self._swaptxcreate_ui_common(t)

	def swaptxcreate5b(self):
		t = self._swaptxcreate(['ETH.MM1', '98.7654321', 'ETH', f'{dfl_sid}:E:12'])
		return self._swaptxcreate_ui_common(t)

	def swaptxsign1(self):
		return self._swaptxsign()

	def swaptxsend1(self):
		return self._swaptxsend()

	def swaptxstatus1(self):
		return self._swaptxsend(add_opts=['--verbose', '--status'], status=True)

	def swaptxmemo4(self):
		import time
		time.sleep(1)
		return self._check_token_swaptx_memo('=:b:mkQsXA7mqDtnUpkaXMbDtAL1KMeof4GPw3:0/3/0')

	def swaptxreceipt4(self):
		return self._swaptxsend(add_opts=['--receipt'], spawn_only=True)

	def swaptxsend5_test(self):
		return self._swaptxsend_eth_proxy(test=True)

	def swaptxsend5a(self):
		return self._swaptxsend_eth_proxy(
			add_opts = ['--txhex-idx=1', f'--proxy=localhost:{TestProxy.port}'])

	def swaptxsend5b(self):
		return self._swaptxsend_eth_proxy(add_opts=['--txhex-idx=2'])

	def swaptxsend5(self):
		return self._swaptxsend_eth_proxy()

	swaptxsign5 = swaptxsign4 = swaptxsign3 = swaptxsign1
	swaptxsend4 = swaptxsend3 = swaptxsend1
	swaptxstatus4 = swaptxstatus1

	def bal1(self):
		return self.bal('swap1')

	def bal2(self):
		return self.bal('swap2')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-passgen: Generate a series or range of passwords from an MMGen
               deterministic wallet
"""

import sys, time

from .cfg import gc, Config
from .addrlist import AddrIdxList
from .passwdlist import PasswordList
from .wallet import Wallet
from .obj import MMGenPWIDString
from .ui import keypress_confirm

pwi = PasswordList.pw_info

opts_data = {
	'sets': [('print_checksum', True, 'quiet', True)],
	'text': {
		'desc': f"""
                 Generate a range or list of passwords from an {gc.proj_name} wallet,
                 mnemonic, seed or brainwallet for the given ID string
		 """,
		'usage':'[opts] [seed source] <ID string> <index list or range(s)>',
		'options': """
-h, --help            Print this help message
--, --longhelp        Print help message for long (global) options
-d, --outdir=      d  Output files to directory 'd' instead of working dir
-e, --echo-passphrase Echo passphrase or mnemonic to screen upon entry
-f, --passwd-fmt=  f  Generate passwords of format 'f'.  Default: {pl.dfl_pw_fmt}.
                      See PASSWORD FORMATS below
-i, --in-fmt=      f  Input is from wallet format 'f' (see FMT CODES below)
-H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
                      'f' at offset 'o' (comma-separated)
-O, --old-incog-fmt   Specify old-format incognito input
-L, --passwd-len=  l  Specify length of generated passwords.  For defaults,
                      see PASSWORD FORMATS below.  An argument of 'h' will
                      generate passwords of half the default length.
-l, --seed-len=    l  Specify wallet seed length of 'l' bits.  This option
                      is required only for brainwallet and incognito inputs
                      with non-standard (< {dsl}-bit) seed lengths.
-p, --hash-preset= p  Use the scrypt hash parameters defined by preset 'p'
                      for password hashing (default: '{gc.dfl_hash_preset}')
-z, --show-hash-presets Show information on available hash presets
-P, --passwd-file= f  Get wallet passphrase from file 'f'
-q, --quiet           Produce quieter output; suppress some warnings
-r, --usr-randchars=n Get 'n' characters of additional randomness from user
                      (min={cfg.min_urandchars}, max={cfg.max_urandchars}, default={cfg.usr_randchars})
-S, --stdout          Print passwords to stdout
-v, --verbose         Produce more verbose output
""",
	'notes': """

                           NOTES FOR THIS COMMAND

ID string must be a valid UTF-8 string not longer than {ml} characters and
not containing the symbols '{fs}'.

Password indexes are given as a comma-separated list and/or hyphen-separated
range(s).

Changing either the password format (base32,base58) or length alters the seed
and thus generates a completely new set of passwords.

PASSWORD FORMATS:

  {pfi}

EXAMPLES:

  Generate ten base58 passwords of length {i58.dfl_len} for Alice's email account:
  {gc.prog_name} alice@nowhere.com 1-10

  Generate ten base58 passwords of length 16 for Alice's email account:
  {gc.prog_name} --passwd-len=16 alice@nowhere.com 1-10

  Generate ten base32 passwords of length {i32.dfl_len} for Alice's email account:
  {gc.prog_name} --passwd-fmt=b32 alice@nowhere.com 1-10

  Generate three BIP39 mnemonic seed phrases of length {i39.dfl_len} for Alice's
  Trezor device:
  {gc.prog_name} --passwd-fmt=bip39 mytrezor 1-3

  All passwords are cryptographically unlinkable with each other, including
  passwords with the same format but different length, so Alice needn't worry
  about inadvertent reuse of private data.


                      NOTES FOR ALL GENERATOR COMMANDS

{n_pw}

{n_bw}

{n_fmt}
"""
	},
	'code': {
		'options': lambda cfg, help_notes, s: s.format(
			cfg   = cfg,
			dsl   = help_notes('dfl_seed_len'),
			pl    = PasswordList,
			gc    = gc,
		),
		'notes': lambda cfg, help_notes, s: s.format(
			cfg   = cfg,
			i58   = pwi['b58'],
			i32   = pwi['b32'],
			i39   = pwi['bip39'],
			ml    = MMGenPWIDString.max_len,
			fs    = '", "'.join(MMGenPWIDString.forbidden),
			n_pw  = help_notes('passwd'),
			n_bw  = help_notes('brainwallet'),
			pfi   = help_notes('password_formats'),
			n_fmt = help_notes('fmt_codes'),
			gc    = gc,
		)
	}
}

cfg = Config(opts_data=opts_data)

if len(cfg._args) < 2:
	cfg._usage()

pw_idxs = AddrIdxList(fmt_str=cfg._args.pop())

pw_id_str = cfg._args.pop()

from .fileutil import get_seed_file
sf = get_seed_file(cfg, nargs=1)

pw_fmt = cfg.passwd_fmt or PasswordList.dfl_pw_fmt
pw_len = pwi[pw_fmt].dfl_len // 2 if cfg.passwd_len in ('h', 'H') else cfg.passwd_len

from .protocol import init_proto
proto = init_proto(cfg, 'btc') # TODO: get rid of dummy proto

PasswordList(
	cfg             = cfg,
	proto           = proto,
	pw_id_str       = pw_id_str,
	pw_len          = pw_len,
	pw_fmt          = pw_fmt,
	chk_params_only = True)

from .ui import do_license_msg
do_license_msg(cfg)

ss = Wallet(cfg, fn=sf)

al = PasswordList(
	cfg       = cfg,
	proto     = proto,
	seed      = ss.seed,
	pw_idxs   = pw_idxs,
	pw_id_str = pw_id_str,
	pw_len    = pw_len,
	pw_fmt    = pw_fmt)

af = al.file

af.format()

if keypress_confirm(cfg, 'Encrypt password list?'):
	af.encrypt()
	af.write(binary=True, desc='encrypted password list')
else:
	if cfg.test_suite_popen_spawn and sys.platform == 'win32':
		time.sleep(0.1)
	af.write(desc='password list')



#!/usr/bin/env python3

"""
test.modtest_d.subseed: subseed unit test for the MMGen suite
"""

from mmgen.util import msg, msg_r

from ..include.common import cfg, vmsg_r

class unit_test:

	def run_test(self, name, ut):
		from mmgen.seed import Seed
		from mmgen.subseed import SubSeedList, SubSeedIdxRange

		nSubseeds = SubSeedList.dfl_len

		def basic_ops():
			msg_r('Testing basic ops...')
			for a, b, c, d, e, f, h in (
					(8, '4710FBF0', '0C1B0615', '803B165C', '2669AC64', 256, '10L'),
					(6, '9D07ABBD', 'EBA9C33F', '20787E6A', '192E2AA2', 192, '10L'),
					(4, '43670520', '04A4CCB3', 'B5F21D7B', 'C1934CFF', 128, '10L'),
				):

				seed_bin = bytes.fromhex('deadbeef' * a)
				seed = Seed(cfg, seed_bin=seed_bin)
				assert seed.sid == b, seed.sid

				subseed = seed.subseed('2s')
				assert subseed.sid == c, subseed.sid

				subseed = seed.subseed('3')
				assert subseed.sid == d, subseed.sid

				subseed = seed.subseed_by_seed_id(e)
				assert subseed.bitlen == f, subseed.bitlen
				assert subseed.sid == e, subseed.sid
				assert subseed.idx == 10, subseed.idx
				assert subseed.ss_idx == h, subseed.ss_idx

				seed2 = Seed(cfg, seed_bin=seed_bin)
				ss2_list = seed2.subseeds

				seed2.subseeds._generate(1)
				assert len(ss2_list) == 1, len(ss2_list)

				seed2.subseeds._generate(1) # do nothing
				seed2.subseeds._generate(2) # append one item

				seed2.subseeds._generate(5)
				assert len(ss2_list) == 5, len(ss2_list)

				seed2.subseeds._generate(3) # do nothing
				assert len(ss2_list) == 5, len(ss2_list)

				seed2.subseeds._generate(10)
				assert len(ss2_list) == 10, len(ss2_list)

				assert seed.pfmt() == seed2.pfmt()
				assert seed.subseeds.pfmt() == seed2.subseeds.pfmt()

				s = seed.subseeds.format(1, nSubseeds)
				s_lines = s.strip().split('\n')
				assert len(s_lines) == nSubseeds + 4, s

				a = seed.subseed('2L').sid
				b = [e for e in s_lines if ' 2L:' in e][0].strip().split()[1]
				assert a == b, b

				c = seed.subseed('2').sid
				assert c == a, c

				a = seed.subseed('5S').sid
				b = [e for e in s_lines if ' 5S:' in e][0].strip().split()[3]
				assert a == b, b

				s = seed.subseeds.format(nSubseeds+1, nSubseeds+2)
				s_lines = s.strip().split('\n')
				assert len(s_lines) == 6, s

				ss_idx = str(nSubseeds+2) + 'S'
				a = seed.subseed(ss_idx).sid
				b = [e for e in s_lines if f' {ss_idx}:' in e][0].strip().split()[3]
				assert a == b, b

				s = seed.subseeds.format(1, 10)
				s_lines = s.strip().split('\n')
				assert len(s_lines) == 14, s

				vmsg_r(f'\n{s}')

			msg('OK')

		def defaults_and_limits():
			msg_r('Testing defaults and limits...')

			seed_bin = bytes.fromhex('deadbeef' * 8)

			seed = Seed(cfg, seed_bin=seed_bin, nSubseeds=11)
			seed.subseeds._generate()
			ss = seed.subseeds
			assert len(ss.data['long']) == len(ss.data['short']), len(ss.data['short'])
			assert len(ss) == 11, len(ss)

			seed = Seed(cfg, seed_bin=seed_bin)
			seed.subseeds._generate()
			ss = seed.subseeds
			assert len(ss.data['long']) == len(ss.data['short']), len(ss.data['short'])
			assert len(ss) == nSubseeds, len(ss)

			seed = Seed(cfg, seed_bin=seed_bin)
			seed.subseed_by_seed_id('EEEEEEEE')
			ss = seed.subseeds
			assert len(ss.data['long']) == len(ss.data['short']), len(ss.data['short'])
			assert len(ss) == nSubseeds, len(ss)

			seed = Seed(cfg, seed_bin=seed_bin)
			subseed = seed.subseed_by_seed_id('803B165C')
			assert len(ss.data['long']) == len(ss.data['short']), len(ss.data['short'])
			assert subseed.sid == '803B165C', subseed.sid
			assert subseed.idx == 3, subseed.idx

			seed = Seed(cfg, seed_bin=seed_bin)
			subseed = seed.subseed_by_seed_id('803B165C', last_idx=1)
			assert len(ss.data['long']) == len(ss.data['short']), len(ss.data['short'])
			assert subseed is None, subseed

			r = SubSeedIdxRange('1-5')
			r2 = SubSeedIdxRange(1, 5)
			assert r2 == r, r2
			assert r == (r.first, r.last), r
			assert r.first == 1, r.first
			assert r.last == 5, r.last
			assert r.items == [1, 2, 3, 4, 5], r.items
			assert list(r.iterate()) == r.items, list(r.iterate())

			r = SubSeedIdxRange('22')
			r2 = SubSeedIdxRange(22, 22)
			assert r2 == r, r2
			assert r == (r.first, r.last), r
			assert r.first == 22, r.first
			assert r.last == 22, r.last
			assert r.items == [22], r
			assert list(r.iterate()) == r.items, list(r.iterate())

			r = SubSeedIdxRange('3-3')
			assert r.items == [3], r.items

			r = SubSeedIdxRange(f'{nSubseeds-1}-{nSubseeds}')
			assert r.items == [nSubseeds-1, nSubseeds], r.items

			for n, e in enumerate(SubSeedIdxRange('1-5').iterate(), 1):
				assert n == e, e

			assert n == 5, n

			msg('OK')

		def collisions():
			ss_count, ltr, last_sid, collisions_chk = (
				(SubSeedIdxRange.max_idx, 'S', '2788F26B', 470),
				(49509, 'L', '8D1FE500', 2)
			)[bool(cfg.fast)]

			last_idx = str(ss_count) + ltr

			msg_r(f'Testing Seed ID collisions ({ss_count} subseed pairs)...')

			seed_bin = bytes.fromhex('12abcdef' * 8) # 95B3D78D
			seed = Seed(cfg, seed_bin=seed_bin)

			seed.subseeds._generate(ss_count)
			ss = seed.subseeds

			assert seed.subseed(last_idx).sid == last_sid, seed.subseed(last_idx).sid

			for sid in ss.data['long']:
				# msg(sid)
				assert sid not in ss.data['short']

			collisions = 0
			for k in ('short', 'long'):
				for sid in ss.data[k]:
					collisions += ss.data[k][sid][1]

			assert collisions == collisions_chk, collisions
			vmsg_r(f'\n{collisions} collisions, last_sid {last_sid}')
			msg('OK')

		basic_ops()
		defaults_and_limits()
		collisions()

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tw.view: Ethereum base protocol base class for tracking wallet view classes
"""

class EthereumTwView:

	def get_disp_prec(self, wide):
		return self.proto.coin_amt.max_prec if wide else 8

	def gen_subheader(self, cw, color):
		if self.disp_prec == 8:
			yield 'Balances truncated to 8 decimal points'
		yield from super().gen_subheader(cw, color)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tw.shared: classes and functions shared by all tracking wallet classes
"""

from ..objmethods import HiliteStr, InitErrors, MMGenObject
from ..obj import TwComment
from ..addr import MMGenID

class TwMMGenID(HiliteStr, InitErrors, MMGenObject):
	color = 'orange'
	width = 0
	trunc_ok = False
	def __new__(cls, proto, id_str):
		if isinstance(id_str, cls):
			return id_str
		try:
			ret = addr = disp = MMGenID(proto, id_str)
			sort_key, idtype = (ret.sort_key, 'mmgen')
		except Exception as e:
			try:
				coin, addr = id_str.split(':', 1)
				assert coin == proto.base_coin.lower(), (
					f'not a string beginning with the prefix {proto.base_coin.lower()!r}:')
				ret, sort_key, idtype, disp = (id_str, 'z_'+id_str, 'non-mmgen', 'non-MMGen')
				addr = proto.coin_addr(addr)
			except Exception as e2:
				return cls.init_fail(e, id_str, e2=e2)

		me = str.__new__(cls, ret)
		me.obj = ret
		me.disp = disp
		me.addr = addr
		me.sort_key = sort_key
		me.type = idtype
		me.proto = proto
		return me

	def fmt(self, width, /, **kwargs):
		return super().fmtc(self.disp, width, **kwargs)

# non-displaying container for TwMMGenID, TwComment
class TwLabel(str, InitErrors, MMGenObject):
	exc = 'BadTwLabel'
	passthru_excs = ('BadTwComment',)
	def __new__(cls, proto, text):
		if isinstance(text, cls):
			return text
		try:
			ts = text.split(None, 1)
			mmid = TwMMGenID(proto, ts[0])
			comment = TwComment(ts[1] if len(ts) == 2 else '')
			me = str.__new__(cls, mmid + (' ' + comment if comment else ''))
			me.mmid = mmid
			me.comment = comment
			me.proto = proto
			return me
		except Exception as e:
			return cls.init_fail(e, text)

def get_tw_label(proto, s):
	"""
	raise an exception on a malformed comment, return None on an empty or invalid label
	"""
	try:
		return TwLabel(proto, s)
	except Exception as e:
		if type(e).__name__ == 'BadTwComment': # do it this way to avoid importing .exception
			raise
		else:
			return None



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test/scrambletest.py: seed scrambling and addrlist data generation tests for all
                      supported coins + passwords
"""

import sys, os, time
from subprocess import run, PIPE
from collections import namedtuple

try:
	from include import test_init
except ImportError:
	from test.include import test_init

from mmgen.main import launch
from mmgen.cfg import Config
from mmgen.util import msg, msg_r, bmsg, die, list_gen
from mmgen.color import gray, green
from test.include.common import set_globals, init_coverage, end_msg

opts_data = {
	'text': {
		'desc': 'Test seed scrambling and addrlist data generation for all supported altcoins',
		'usage':'[options] [command]',
		'options': """
-h, --help          Print this help message
--, --longhelp      Print help message for long (global) options
-a, --no-altcoin    Skip altcoin tests
-C, --coverage      Produce code coverage info using trace module
-l, --list-cmds     List and describe the tests and commands in this test suite
-v, --verbose       Produce more verbose output
""",
	'notes': """
Valid commands: 'coin', 'pw'
If no command is given, the whole suite of tests is run.
"""
	}
}

cfg = Config(opts_data=opts_data)

set_globals(cfg)

td = namedtuple('scrambletest_entry', ['seed', 'str', 'id_str', 'lbl', 'addr'])

bitcoin_data = {
#                   SCRAMBLED_SEED[:8] SCRAMBLE_KEY     ID_STR   LBL          FIRST ADDR
'btc':           td('456d7f5f1c4bfe3b','(none)',        '',      '',          '1MU7EdgqYy9JX35L25hR6CmXXcSEBDAwyv'),
'btc_compressed':td('bf98a4af5464a4ef','compressed',    '-C',    'COMPRESSED','1F97Jd89wwmu4ELadesAdGDzg3d8Y6j5iP'),
'btc_segwit':    td('b56962d829ffc678','segwit',        '-S',    'SEGWIT',    '36TvVzU5mxSjJ3D9qKAmYzCV7iUqtTDezF'),
'btc_bech32':    td('d09eea818f9ad17f','bech32',        '-B',    'BECH32',    'bc1q8snv94j6959y3gkqv4gku0cm5mersnpucsvw5z'),
}

altcoin_data = {
'bch':           td('456d7f5f1c4bfe3b','(none)',        '',      '',              '1MU7EdgqYy9JX35L25hR6CmXXcSEBDAwyv'),
'bch_compressed':td('bf98a4af5464a4ef','compressed',    '-C',    'COMPRESSED',    '1F97Jd89wwmu4ELadesAdGDzg3d8Y6j5iP'),
'ltc':           td('b11f16632e63ba92','ltc:legacy',    '-LTC',  'LTC',           'LMxB474SVfxeYdqxNrM1WZDZMnifteSMv1'),
'ltc_compressed':td('7ccf465d466ee7d3','ltc:compressed','-LTC-C','LTC:COMPRESSED','LdkebBKVXSs6NNoPJWGM8KciDnL8LhXXjb'),
'ltc_segwit':    td('9460f5ba15e82768','ltc:segwit',    '-LTC-S','LTC:SEGWIT',    'MQrY3vEbqKMBgegXrSaR93R2HoTDE5bKrY'),
'ltc_bech32':    td('dbdbff2e196e27d3','ltc:bech32',    '-LTC-B','LTC:BECH32',    'ltc1qdvgqsz94ht20lr8fyk5v7n884hu9p7d8k9easu'),
'eth':           td('213ed116869b19f2','eth',           '-ETH',  'ETH',           'e704b6cfd9f0edb2e6cfbd0c913438d37ede7b35'),
'etc':           td('909def37096f5ab8','etc',           '-ETC',  'ETC',           '1a6acbef8c38f52f20d04ecded2992b04d8608d7'),
'dash':          td('1319d347b021f952','dash:legacy',   '-DASH', 'DASH',          'XoK491fppGNZQUUS9uEFkT6L9u8xxVFJNJ'),
'emc':           td('7e1a29853d2db875','emc:legacy',    '-EMC',  'EMC',           'EU4L6x2b5QUb2gRQsBAAuB8TuPEwUxCNZU'),
'zec':           td('0bf9b5b20af7b5a0','zec:legacy',    '-ZEC',  'ZEC',           't1URz8BHxV38v3gsaN6oHQNKC16s35R9WkY'),
'zec_zcash_z':   td('b15570d033df9b1a','zec:zcash_z',   '-ZEC-Z','ZEC:ZCASH_Z',   'zcLMMsnzfFYZWU4avRBnuc83yh4jTtJXbtP32uWrs3ickzu1krMU4ppZCQPTwwfE9hLnRuFDSYF8VFW13aT9eeQK8aov3Ge'),
'xmr':           td('c76af3b088da3364','xmr:monero',    '-XMR-M','XMR:MONERO',    '41tmwZd2CdXEGtWqGY9fH9FVtQM8VxZASYPQ3VJQhFjtGWYzQFuidD21vJYTi2yy3tXRYXTNXBTaYVLav62rwUUpFFyicZU'),
}

passwd_data = {
'dfl_dfl_αω':  td('b78c363f3d6714f6', 'b58:20:αω',  '-αω-b58-20',  'αω b58:20',  'L3JcAZze6LQS2TFodoW9'),
'dfl_H_αω':    td('2cba1ba1a73d5497', 'b58:10:αω',  '-αω-b58-10',  'αω b58:10',  'KgSBfzPBWj'),
'b32_dfl_αω':  td('ad71ff8d0512660d', 'b32:24:αω',  '-αω-b32-24',  'αω b32:24',  'VQFESDGWAQCOCSG6GDLIG5OQ'),
'b32_H_αω':    td('b050a1613dd4d3ad', 'b32:12:αω',  '-αω-b32-12',  'αω b32:12',  'QTI7HTVN3JOE'),
'hex_dfl_αω':  td('8322b26abd931d55', 'hex:64:αω',  '-αω-hex-64',  'αω hex:64',  '6e9f8d5d2ec6c4f2a079b163e37f5dc2a7e6adb221c4de315078ffad971ba260'),
'hex_H_αω':    td('46beb852d38ed5c4', 'hex:32:αω',  '-αω-hex-32',  'αω hex:32',  '86143b36b649728cc8f3677872ed37d3'),
'bip39_dfl_αω':td('95b383d5092a55df', 'bip39:24:αω','-αω-bip39-24','αω bip39:24','treat athlete brand top beauty poverty senior unhappy vacant domain yellow scale fossil aim lonely fatal sun nuclear such ancient stage require stool similar'),
'bip39_18_αω': td('29e5a605ffa36142', 'bip39:18:αω','-αω-bip39-18','αω bip39:18','better legal various ketchup then range festival either tomato cradle say absorb solar earth alter pattern canyon liar'),
'bip39_12_αω': td('efa13cb309d7fc1d', 'bip39:12:αω','-αω-bip39-12','αω bip39:12','lady oppose theme fit position merry reopen acquire tuna dentist young chunk'),
'xmrseed_dfl_αω':td('62f5b72a5ca89cab', 'xmrseed:25:αω','-αω-xmrseed-25','αω xmrseed:25','tequila eden skulls giving jester hospital dreams bakery adjust nanny cactus inwardly films amply nanny soggy vials muppet yellow woken ashtray organs exhale foes eden'),
}

run_env = dict(os.environ)
run_env['MMGEN_DEBUG_ADDRLIST'] = '1'
words_file = 'test/ref/98831F3A.mmwords'
cvrg_opts = '-m trace --count --coverdir={} --file={}'.format(*init_coverage()).split() if cfg.coverage else []

def make_cmd(progname, opts, add_opts, args):
	return ['python3'] + cvrg_opts + [f'cmds/{progname}', '-qS'] + opts + add_opts + [words_file] + args + ['1']

def run_cmd(cmd):
	cp = run(cmd, stdout=PIPE, stderr=PIPE, text=True, env=run_env)
	if cp.returncode != 0:
		die(2, f'\nSpawned program exited with error code {cp.returncode}:\n{cp.stderr}')
	return cp.stdout.splitlines()

def run_test(progname, opts, add_opts, args, test_data, addr_desc, opts_w):
	cmd = make_cmd(progname, opts, add_opts, args)
	if cfg.verbose:
		msg(green(f'Executing: {" ".join(cmd)}'))
	else:
		msg_r('Testing: {} {:{w}} '.format(progname, ' '.join(opts), w=opts_w))

	lines = run_cmd(cmd)
	cmd_out = dict([e[9:].split(': ') for e in lines if e.startswith('sc_debug_')])
	cmd_out['addr'] = lines[-2].split(None, 1)[-1]

	ref_data = test_data._asdict()
	for k in ref_data:
		if cmd_out[k] == ref_data[k]:
			s = k.replace('seed', 'seed[:8]').replace('addr', addr_desc)
			cfg._util.vmsg(f'  {s:9}: {cmd_out[k]}')
		else:
			die(4, f'\nError: sc_{k} value {cmd_out[k]} does not match reference value {ref_data[k]}')
	msg(green('OK') if cfg.verbose else 'OK')

def make_coin_test_data():
	bmsg('Testing address scramble strings and list IDs')
	coin_data = bitcoin_data | ({} if cfg.no_altcoin else altcoin_data)
	for id_str, test_data in coin_data.items():
		if id_str == 'zec_zcash_z' and sys.platform == 'win32':
			msg(gray("Skipping 'zec_zcash_z' test for Windows platform"))
			continue
		coin, mmtype = id_str.split('_', 1) if '_' in id_str else (id_str, None)
		opts = list_gen(
			[f'--coin={coin}'],
			[f'--type={mmtype}', mmtype],
			['--cashaddr=0', coin == 'bch']
		)
		yield ('mmgen-addrgen', opts, [], [], test_data, 'address')

def make_passwd_test_data():
	bmsg('Testing password scramble strings and list IDs')
	for id_str, test_data in passwd_data.items():
		if id_str.startswith('xmrseed') and cfg.no_altcoin:
			continue
		pw_fmt, pw_len, pw_id = id_str.split('_')
		opts = list_gen(
			[f'--passwd-fmt={pw_fmt}', pw_fmt != 'dfl'],
			[f'--passwd-len={pw_len}', pw_len != 'dfl'],
		)
		yield ('mmgen-passgen', opts, ['--accept-defaults'], [pw_id], test_data, 'password')

def main():
	start_time = int(time.time())

	cmds = cfg._args or ('coin', 'pw')
	funcs = {'coin': make_coin_test_data, 'pw': make_passwd_test_data}
	for cmd in cmds:
		data = tuple(funcs[cmd]())
		opts_w = max(len(' '.join(e[1])) for e in data)
		for d in data:
			run_test(*d, opts_w)

	end_msg(int(time.time()) - start_time)

launch(func=main)



#!/usr/bin/env python3

import os

from mmgen.cfg import Config
from mmgen.util import msg

opts_data = {
	'sets': [
		('no_license', True, 'grokify', True),
		('autosign', True, 'foobleize', True)
	],
	'text': {
		'desc': 'Cfg file test',
		'usage':'[args] [opts]',
		'options': """
-a, --autosign   Turn on autosigning
-F, --foobleize  Foobleize the input
-G, --grokify    Grokify the output
"""},
}

cfg = Config(opts_data=opts_data)

cmd_args = cfg._args
op, args = (cmd_args[0], cmd_args[1:]) if cmd_args else (None, None)

from mmgen.cfgfile import mmgen_cfg_file

cf_usr = mmgen_cfg_file(cfg, 'usr')
cf_sys = mmgen_cfg_file(cfg, 'sys')
cf_sample = mmgen_cfg_file(cfg, 'sample')

msg(f'Usr cfg file:    {os.path.relpath(cf_usr.fn)}')
msg(f'Sys cfg file:    {os.path.relpath(cf_sys.fn)}')
msg(f'Sample cfg file: {os.path.relpath(cf_sample.fn)}')

if op:
	if op == 'print_cfg':
		for name in args:
			msg('{} {}'.format(name+':', getattr(cfg, name)))
	elif op == 'parse_test':
		ps = cf_sample.get_lines()
		msg(f'parsed chunks: {len(ps)}')
		pu = cf_usr.get_lines()
		msg('usr cfg: {}'.format(' '.join(f'{i.name}={i.value}' for i in pu)))
	elif op == 'coin_specific_vars':
		for varname in args:
			msg('{}.{}.{}: {}'.format(
				cfg._proto.coin.lower(),
				type(cfg._proto).__name__,
				varname,
				getattr(cfg._proto, varname)
			))
	elif op == 'autoset_opts':
		assert cfg.rpc_backend == 'aiohttp', "cfg.rpc_backend != 'aiohttp'"
	elif op == 'autoset_opts_cmdline':
		assert cfg.rpc_backend == 'curl', "cfg.rpc_backend != 'curl'"
	elif op == 'mnemonic_entry_modes':
		from mmgen.mn_entry import mn_entry
		msg('mnemonic_entry_modes: {}\nmmgen: {}\nbip39: {}'.format(
			cfg.mnemonic_entry_modes,
			mn_entry(cfg, 'mmgen').usr_dfl_entry_mode,
			mn_entry(cfg, 'bip39').usr_dfl_entry_mode))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-regtest: Coin daemon regression test mode setup and operations for the MMGen
               suite
"""

from .cfg import gc, Config
from .util import die, async_run

opts_data = {
	'sets': [('yes', True, 'quiet', True)],
	'text': {
		'desc': f'Coin daemon regression test mode setup and operations for the {gc.proj_name} suite',
		'usage':   '[opts] <command>',
		'options': """
-h, --help          Print this help message
--, --longhelp      Print help message for long (global) options
-b, --bdb-wallet    Create and use a legacy Berkeley DB coin daemon wallet
-e, --empty         Don't fund Bob and Alice's wallets on setup
-n, --setup-no-stop-daemon  Don't stop daemon after setup is finished
-q, --quiet         Produce quieter output
-v, --verbose       Produce more verbose output
""",
	'notes': """

                         AVAILABLE COMMANDS

  setup           - set up Bob and Alice regtest mode
  start           - start the regtest coin daemon
  stop            - stop the regtest coin daemon
  generate N      - mine N blocks (defaults to 1)
  send ADDR AMT   - send amount AMT of miner funds to address ADDR
  state           - show current state of daemon (ready, busy, or stopped)
  balances        - get Bob and Alice's balances
  mempool         - show transaction IDs in mempool
  cli             - execute an RPC call with supplied arguments
  wallet_cli      - execute a wallet RPC call with supplied arguments (wallet
                    is first argument)
	"""
	}
}

cfg = Config(opts_data=opts_data)

cmd_args = cfg._args

from .proto.btc.regtest import MMGenRegtest

def check_num_args():
	m = getattr(MMGenRegtest, cmd_args[0])
	margs = m.__code__.co_varnames[1:m.__code__.co_argcount]
	mdfls = m.__defaults__ or ()
	amin = len(margs) - len(mdfls)
	amax = len(margs)
	args = cmd_args[1:]
	m = "{}: too {} arguments for command '%s' (must have no {} than {})" % cmd_args[0]
	if len(args) < amin:
		die(1, m.format(args, 'few', 'less', amin))
	elif len(cmd_args[1:]) > amax:
		die(1, m.format(args, 'many', 'more', amax))

if not cmd_args:
	cfg._usage()
elif cmd_args[0] not in MMGenRegtest.usr_cmds:
	die(1, f'{cmd_args[0]!r}: invalid command')
elif cmd_args[0] not in ('cli', 'wallet_cli', 'balances'):
	check_num_args()

async def main():
	await MMGenRegtest(cfg, cfg.coin, bdb_wallet=cfg.bdb_wallet).cmd(cmd_args)

async_run(main())
